"""
This module contains predefined basic pulsed experiments inheriting from the PulsedSim class as part of the QuaCCAToo package.

Classes
-------
- Rabi: resonant pulse of varying duration, such that the quantum system will undergo periodical transitions between the excited and ground states.
- PMR: Pulsed Magnetic Resonance (PMR) experiment, composed by a single pulse where the frequency is changed such that when it corresponds to a transition in the Hamiltonian of the system, the observable will be affected.
- Ramsey: Ramsey experiment, consisting of a free evolution that causes a phase accumulation between states in the system which can be used for interferometry.
- Hahn: Hahn echo experiment, consisting of two free evolutions with a pi pulse in the middle, in order to cancel out dephasings. The Hahn echo is usually used to measure the coherence time of a quantum system, however it can also be used to sense coupled spins.
"""

import numpy as np
from qutip import Qobj, mesolve, propagator

from .pulse_shapes import square_pulse
from .pulsed_sim import PulsedSim

####################################################################################################


class Rabi(PulsedSim):
    """
    A class containing Rabi experiments.

    A Rabi sequence is composed of a resonant pulse of varying duration,
    such that the quantum system will undergo periodical transitions between the excited and ground states.

    Methods
    -------
    run :
        Runs the simulation and stores the results in the results attribute.

    Notes
    -----
    The Rabi sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self, pulse_duration, system, H1, H2=None, pulse_shape=square_pulse, pulse_params=None, options=None
    ):
        """
        Constructor for the Rabi pulsed experiment class.

        Parameters
        ----------
        pulse_duration : numpy.ndarray
            Time array for the simulation representing the pulse duration to be used as the variable for the simulation.
        system : QSys
            Quantum system object containing the initial state, internal Hamiltonian and collapse operators.
        H1 : Qobj or list(Qobj)
            Control Hamiltonian of the system.
        H2 : Qobj or list(Qobj)
            Time-dependent sensing Hamiltonian of the system.
        pulse_shape : FunctionType or list(FunctionType)
            Pulse shape function or list of pulse shape functions representing the time modulation of H1.
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions.
        options : dict
            Dictionary of solver options from Qutip.
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1, pulse_shape, pulse_params, options, len(pulse_duration), None, None, None, None
        )

        # check whether pulse_duration is a numpy array and if it is, assign it to the object
        if (
            not isinstance(pulse_duration, (np.ndarray, list))
            or not np.all(np.isreal(pulse_duration))
            or not np.all(np.greater_equal(pulse_duration, 0))
        ):
            raise ValueError("pulse_duration must be a numpy array of real positive elements")
        else:
            self.total_time = pulse_duration[-1]
            self.variable = pulse_duration
            self.variable_name = f"Pulse Duration (1/{self.system.units_H0})"

        if isinstance(H1, Qobj):
            self.pulse_profiles = [[H1, pulse_duration, pulse_shape, self.pulse_params]]
        else:
            self.pulse_profiles = [
                [H1[i], pulse_duration, pulse_shape[i], self.pulse_params] for i in range(len(H1))
            ]

    def run(self):
        """
        Overwrites the run method of the parent class. Runs the simulation and stores the results in the results attribute.
        If the system has no initial state, the propagator is calcualated.
        If an observable is given, the expectation values are stored in the results attribute.
        For the Rabi sequence, the calculation is optimally performed sequentially instead of in parallel over the pulse lengths,
        thus the run method from the parent class is overwritten.
        """
        if self.system.rho0 is None:
            self.U = propagator(
                self.Ht,
                2 * np.pi * self.variable,
                self.system.c_ops,
                options=self.options,
                args=self.pulse_params,
            )
        else:
            # calculates the density matrices in sequence using mesolve
            self.rho = mesolve(
                self.Ht,
                self.system.rho0,
                2 * np.pi * self.variable,
                self.system.c_ops,
                e_ops=[],
                options=self.options,
                args=self.pulse_params,
            ).states
            self._get_results()


####################################################################################################


class PMR(PulsedSim):
    """
    A class containing Pulsed Magnetic Resonance (PMR) experiments where the frequency is the variable being changed.

    The PMR consists of a single pulse of fixed length and changing frequency. If the frequency matches a resonance of the system,
    it will undergo some transition which will affect the observable.
    This way, the differences between energy levels can be determined with the linewidth usually limited by the pulse length.
    Here we make reference to optical detection as it is the most common detection scheme of pulsed magnetic resonance in color centers,
    however the method can be more general.

    Attributes
    ----------
    frequencies : numpy.ndarray
        Array of frequencies to run the simulation.
    pulse_duration : float or int
        Duration of the pulse.

    Methods
    -------
    PMR_sequence :
        Defines the Pulsed Magnetic Resonance (PMR) sequence for a given frequency of the pulse.
        To be called by the parallel_map in run method.
    plot_pulses :
        Overwrites the plot_pulses method of the parent class in order to first define a pulse frequency to be plotted.

    Notes
    -----
    The PMR sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self,
        frequencies,
        system,
        pulse_duration,
        H1,
        H2=None,
        pulse_shape=square_pulse,
        pulse_params=None,
        time_steps=100,
        options=None,
    ):
        """
        Constructor for the PMR pulsed experiment class.

        Parameters
        ----------
        frequencies : numpy.ndarray
            Array of frequencies to run the simulation.
        system : QSys
            Quantum system object containing the initial state, internal Hamiltonian and collapse operators.
        pulse_duration : float or int
            Duration of the pulse.
        H1 : Qobj or list(Qobj)
            Control Hamiltonian of the system.
        H2 : Qobj or list(Qobj)
            Time-dependent sensing Hamiltonian of the system.
        pulse_shape : FunctionType or list(FunctionType)
            Pulse shape function or list of pulse shape functions representing the time modulation of H1.
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions.
        time_steps : int
            Number of time steps in the pulses for the simulation.
        options : dict
            Dictionary of solver options from Qutip.
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1, pulse_shape, pulse_params, options, time_steps, None, None, None, None
        )

        # check whether frequencies is a numpy array or list and if it is, assign it to the object
        if (
            not isinstance(frequencies, (np.ndarray, list))
            or not np.all(np.isreal(frequencies))
            or not np.all(np.greater_equal(frequencies, 0))
        ):
            raise ValueError("frequencies must be a numpy array or list of real positive elements")
        else:
            self.variable = frequencies
            self.variable_name = f"Frequency ({self.system.units_H0})"

        # check whether pulse_duration is a numpy array and if it is, assign it to the object
        if not isinstance(pulse_duration, (float, int)) or pulse_duration <= 0:
            raise ValueError("pulse_duration must be a positive real number")
        else:
            self.pulse_duration = pulse_duration

        # set the sequence attribute to the PMR_sequence method
        self.sequence = self.PMR_sequence

    def PMR_sequence(self, f):
        """
        Defines the Pulsed Magnetic Resonance (PMR) sequence for a given frequency of the pulse.
        To be called by the parallel_map in run method.

        Parameters
        ----------
        f : float
            Frequency of the pulse.

        Returns
        -------
        rho : Qobj
            Final state.
        """
        self.pulse_params["f_pulse"] = f

        self._pulse(self.Ht, self.pulse_duration, self.options, self.pulse_params)

        return self.rho

    def plot_pulses(
        self, figsize=(6, 4), xlabel="Time", ylabel="Pulse Intensity", title="Pulse Profiles", f_pulse=None
    ):
        """
        Overwrites the plot_pulses method of the parent class in order to first define a pulse frequency to be plotted.

        Parameters
        ----------
        f_pulse : float or int
            Frequency of the pulse to be plotted.
        """
        # if f_pulse is None, assign the first element of the variable attribute to the pulse_params dictionary
        if f_pulse is None:
            self.pulse_params["f_pulse"] = self.variable[0]
        # if f_pulse is a float or an integer, assign it to the pulse_params dictionary
        elif isinstance(f_pulse, (int, float)):
            self.pulse_params["f_pulse"] = f_pulse
        else:
            raise ValueError("f_pulse must be a float or an integer")

        self.total_time = self.pulse_duration

        super().plot_pulses(figsize, xlabel, ylabel, title)


####################################################################################################


class Ramsey(PulsedSim):
    """
    A class containing Ramsey experiments.

    The Ramsey sequence consists of a free evolution in the plane perpendicular to the quantization axis,
    which causes a phase accumulation between states in the system which can be used for sensing.

    Methods
    -------
    ramsey_sequence :
        Defines the Ramsey sequence for a given free evolution time tau and the set of attributes defined in the constructor.
        The sequence consists of an initial pi/2 pulse and a single free evolution.
        The sequence is to be called by the parallel_map method of QuTip.
    _get_pulse_profiles :
        Generates the pulse profiles for the Ramsey sequence for a given tau.
        The pulse profiles are stored in the pulse_profiles attribute of the object.
    plot_pulses :
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the Ramsey sequence for a given tau and then plot them.

    Notes
    -----
    The Ramsey sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self,
        free_duration,
        system,
        pi_pulse_duration,
        H1,
        H2=None,
        projection_pulse=True,
        pulse_shape=square_pulse,
        pulse_params=None,
        options=None,
        time_steps=100,
    ):
        """
        Class constructor for the Ramsey pulsed experiment class.

        Parameters
        ----------
        free_duration : numpy.ndarray
            Time array for the simulation representing the free evolution time to be used as the variable attribute for the simulation.
        system : QSys
            Quantum system object containing the initial state, internal Hamiltonian and collapse operators.
        H1 : Qobj or list(Qobj)
            Control Hamiltonian of the system.
        pi_pulse_duration : float or int
            Duration of the pi pulse.
        H2 : Qobj or list(Qobj)
            Time-dependent sensing Hamiltonian of the system.
        projection_pulse : bool
            Boolean to determine if the measurement is to be performed in the Sz basis or not.
            If True, a final pi/2 pulse is included in order to project the result into the Sz basis, as for most color centers.
        pulse_shape : FunctionType or list(FunctionType)
            Pulse shape function or list of pulse shape functions representing the time modulation of H1.
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions.
        time_steps : int
            Number of time steps in the pulses for the simulation.
        options : dict
            Dictionary of solver options from Qutip.
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1,
            pulse_shape,
            pulse_params,
            options,
            time_steps,
            free_duration,
            pi_pulse_duration,
            None,
            projection_pulse,
        )
        self.sequence = self.ramsey_sequence

    def ramsey_sequence(self, tau):
        """
        Defines the Ramsey sequence for a given free evolution time tau and the set of attributes defined in the constructor.
        The sequence consists of an initial pi/2 pulse and a single free evolution.
        If the projection_pulse is set to True, a final pi/2 pulse is included in order to project the result into the Sz basis.
        The sequence is to be called by the parallel_map method of QuTip.

        Parameters
        ----------
        tau : float
            Free evolution time.

        Returns
        -------
        rho : Qobj
            Final state.
        """

        self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params)
        self._free_evolution(tau - self.pi_pulse_duration / 2, self.options)

        if self.projection_pulse:
            self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params)

        return self.rho

    def _get_pulse_profiles(self, tau=None):
        """
        Generates the pulse profiles for the Ramsey sequence for a given tau.
        The pulse profiles are stored in the pulse_profiles attribute of the object.

        Parameters
        ----------
        tau : float
            Free evolution variable or pulse spacing for the Hahn echo sequence.
        """
        # check if tau is correctly defined
        if tau is None:
            tau = self.variable[-1]
        elif not isinstance(tau, (int, float)) or tau < self.pi_pulse_duration:
            raise ValueError("tau must be a positive real number larger than pi_pulse_duration")

        self.pulse_profiles = []

        # if projection_pulse is True, include the final pi/2 pulse in the pulse_profiles
        if self.projection_pulse:
            ps = tau - self.pi_pulse_duration

            # if only one control Hamiltonian is given, append the pulse_profiles with the Ramsey sequence
            if isinstance(self.H1, Qobj):
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params,
                    ]
                )
                t0 = self.pi_pulse_duration / 2
                self.pulse_profiles.append([None, [t0, ps + t0], None, None])
                t0 += ps
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params,
                    ]
                )
                t0 += self.pi_pulse_duration / 2

            # otherwise if a list of control Hamiltonians is given, it sums over all H1 and appends to the pulse_profiles
            elif isinstance(self.H1, list):
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params,
                        ]
                        for i in range(len(self.H1))
                    ]
                )
                t0 = self.pi_pulse_duration / 2
                self.pulse_profiles.append([None, [t0, ps + t0], None, None])
                t0 += ps
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params,
                        ]
                        for i in range(len(self.H1))
                    ]
                )
                t0 += self.pi_pulse_duration / 2

        # if projection_pulse is false, do not include the final pi/2 pulse in the pulse_profiles
        else:
            ps = tau - self.pi_pulse_duration / 2

            # if only one control Hamiltonian is given, append the pulse_profiles with the Ramsey sequence
            if isinstance(self.H1, Qobj):
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params,
                    ]
                )
                t0 = self.pi_pulse_duration / 2
                self.pulse_profiles.append([None, [t0, ps + t0], None, None])
                t0 += ps

            # otherwise if a list of control Hamiltonians is given, it sums over all H1 and appends to the pulse_profiles
            elif isinstance(self.H1, list):
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params,
                        ]
                        for i in range(len(self.H1))
                    ]
                )
                t0 = self.pi_pulse_duration / 2
                self.pulse_profiles.append([None, [t0, ps + t0], None, None])
                t0 += ps

        self.total_time = t0

    def plot_pulses(
        self,
        figsize=(6, 4),
        xlabel="Time",
        ylabel="Pulse Intensity",
        title="Pulse Profiles of Ramsey Sequence",
        tau=None,
    ):
        """
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the Ramsey sequence for a given tau and then plot them.

        Parameters
        ----------
        tau : float
            Free evolution time for the Hahn echo sequence. Contrary to the run method, the free evolution must be a single number in order to plot the pulse profiles.
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot.
        xlabel : str
            Label of the x-axis.
        ylabel : str
            Label of the y-axis.
        title : str
            Title of the plot.
        """
        # generate the pulse profiles for the Ramsey sequence for a given tau
        self._get_pulse_profiles(tau)

        # call the plot_pulses method of the parent class
        super().plot_pulses(figsize, xlabel, ylabel, title)


####################################################################################################


class Hahn(PulsedSim):
    """
    A class containing Hahn echo experiments.

    The Hahn echo sequence consists of two free evolutions with a pi pulse in the middle, in order to cancel out dephasings.
    The Hahn echo is usually used to measure the coherence time of a quantum system, however it can also be used to sense coupled spins.

    Methods
    -------
    hahn_sequence :
        Defines the Hahn echo sequence for a given free evolution time tau and the set of attributes defined in the constructor,
        returning the final state. The sequence is to be called by the parallel_map method of QuTip.
    _get_pulse_profiles :
        Generates the pulse profiles for the Hahn echo sequence for a given tau.
        The pulse profiles are stored in the pulse_profiles attribute of the object.
    plot_pulses :
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the Hahn echo sequence for a given tau and then plot them.

    Notes
    -----
    The Hahn echo sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self,
        free_duration,
        system,
        pi_pulse_duration,
        H1,
        H2=None,
        projection_pulse=True,
        pulse_shape=square_pulse,
        pulse_params=None,
        options=None,
        time_steps=100,
    ):
        """
        Constructor for the Hahn echo pulsed experiment class, taking a specific free_duration to run the simulation.

        Parameters
        ----------
        free_duration : numpy.ndarray
            Time array for the simulation representing the free evolution time to be used as the variable attribute for the simulation.
        system : QSys
            Quantum system object containing the initial state, internal Hamiltonian and collapse operators.
        H1 : Qobj or list of Qobj
            Control Hamiltonian of the system.
        pi_pulse_duration : float or int
            Duration of the pi pulse.
        H2 : Qobj or list of Qobj
            Time dependent sensing Hamiltonian of the system.
        projection_pulse : bool
            Boolean to determine if the measurement is to be performed in the Sz basis or not.
            If True, a final pi/2 pulse is included in order to project the result into the Sz basis, as done for the most color centers.
        pulse_shape : FunctionType or list of FunctionType
            Pulse shape function or list of pulse shape functions representing the time modulation of H1.
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions.
        time_steps : int
            Number of time steps in the pulses for the simulation.
        options : dict
            Dictionary of solver options from Qutip.
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1,
            pulse_shape,
            pulse_params,
            options,
            time_steps,
            free_duration,
            pi_pulse_duration,
            None,
            projection_pulse,
        )
        self.sequence = self.hahn_sequence

    def hahn_sequence(self, tau):
        """
        Defines the Hahn echo sequence for a given free evolution time tau and the set of attributes defined in the constructor.

        The sequence consists of a pi/2 pulse, a free evolution time tau, a pi pulse and another free evolution time tau followed by a pi/2 pulse, if the projection pulse is set to True.
        The sequence is to be called by the parallel_map method of QuTip.

        Parameters
        ----------
        tau : float
            Free evolution time.

        Returns
        -------
        rho : Qobj
            Final state.
        """
        # pulse separation time
        ps = tau - self.pi_pulse_duration

        self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params)
        self._free_evolution(ps, self.options)
        self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params)

        if self.projection_pulse:
            self._free_evolution(ps, self.options)
            self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params)
        else:
            self._free_evolution(ps + self.pi_pulse_duration / 2, self.options)

        return self.rho

    def _get_pulse_profiles(self, tau=None):
        """
        Generates the pulse profiles for the Hahn echo sequence for a given tau.
        The pulse profiles are stored in the pulse_profiles attribute of the object.

        Parameters
        ----------
        tau : float
            Free evolution variable or pulse spacing for the Hahn echo sequence.
        """
        # check if tau is correctly defined
        if tau is None:
            tau = self.variable[-1]
        elif not isinstance(tau, (int, float)) or tau < self.pi_pulse_duration:
            raise ValueError("tau must be a positive real number larger than pi_pulse_duration")

        self.pulse_profiles = []

        # if projection_pulse is True, include the final pi/2 pulse in the pulse_profiles
        if self.projection_pulse:
            # if only one control Hamiltonian is given, append the pulse_profiles with the Hahn echo sequence as in the hahn_sequence method
            if isinstance(self.H1, Qobj):
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params,
                    ]
                )
                t0 = self.pi_pulse_duration / 2
                self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration], None, None])
                t0 += tau - self.pi_pulse_duration
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params,
                    ]
                )
                t0 += self.pi_pulse_duration
                self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration], None, None])
                t0 += tau - self.pi_pulse_duration
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params,
                    ]
                )
                t0 += self.pi_pulse_duration / 2

            # otherwise if a list of control Hamiltonians is given, it sums over all H1 and appends to the pulse_profiles the Hahn echo sequence as in the hahn_sequence method
            elif isinstance(self.H1, list):
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params,
                        ]
                        for i in range(len(self.H1))
                    ]
                )
                t0 = self.pi_pulse_duration / 2
                self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration], None, None])
                t0 += tau - self.pi_pulse_duration
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params,
                        ]
                        for i in range(len(self.H1))
                    ]
                )
                t0 += self.pi_pulse_duration
                self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration], None, None])
                t0 += tau - self.pi_pulse_duration
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params,
                        ]
                        for i in range(len(self.H1))
                    ]
                )
                t0 += self.pi_pulse_duration / 2

        # if projection_pulse is False, do not include the final pi/2 pulse in the pulse_profiles
        elif isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params,
                ]
            )
            t0 = self.pi_pulse_duration / 2
            self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration], None, None])
            t0 += tau - self.pi_pulse_duration
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params,
                ]
            )
            t0 += self.pi_pulse_duration
            self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration / 2], None, None])
            t0 += tau - self.pi_pulse_duration / 2

        # otherwise if a list of control Hamiltonians is given, it sums over all H1 and appends to the pulse_profiles the Hahn echo sequence as in the hahn_sequence method
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params,
                    ]
                    for i in range(len(self.H1))
                ]
            )
            t0 = self.pi_pulse_duration / 2
            self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration], None, None])
            t0 += tau - self.pi_pulse_duration
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params,
                    ]
                    for i in range(len(self.H1))
                ]
            )
            t0 += self.pi_pulse_duration
            self.pulse_profiles.append([None, [t0, t0 + tau - self.pi_pulse_duration / 2], None, None])
            t0 += tau - self.pi_pulse_duration / 2

        # set the total_time attribute to the total time of the pulse sequence
        self.total_time = t0

    def plot_pulses(
        self, figsize=(6, 6), xlabel="Time", ylabel="Pulse Intensity", title="Pulse Profiles", tau=None
    ):
        """
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the Hahn echo sequence for a given tau and then plot them.

        Parameters
        ----------
        tau : float
            Free evolution time for the Hahn echo sequence. Contrary to the run method, the free evolution must be a single number in order to plot the pulse profiles.
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot.
        xlabel : str
            Label of the x-axis.
        ylabel : str
            Label of the y-axis.
        title : str
            Title of the plot.
        """
        # generate the pulse profiles for the Hahn echo sequence for a given tau
        self._get_pulse_profiles(tau)

        # call the plot_pulses method of the parent class
        super().plot_pulses(figsize, xlabel, ylabel, title)



"""
This module provides functions to save and load quaccatoo objects, such as instances from QSys, ExpData, Analysis and PulsedSim.
"""

import inspect
import os
import shutil
import zipfile

import dill
import numpy as np
from qutip import Qobj, fileio

import quaccatoo


def save(object, file_name):
    """
    Look for all the attributes of the object, save Qobj attributes to separate files,
    and save the rest of the attributes to a pickle file. Finally, create a zip file
    containing all the files.

    Parameters
    ----------
    object : quaccatoo object
        The object to be saved.
    file_name : str
        Path to the file where the attributes will be saved.
    """
    if not isinstance(file_name, str):
        raise ValueError("file_name must be a string")

    # create a temporary directory to store files
    tmp_dir = "tmp"
    os.makedirs(tmp_dir, exist_ok=True)

    try:
        # Get a list of all the attributes defined in the object
        attributes = [
            attr
            for attr in dir(object)
            if not attr.startswith("__")
            and not (inspect.isfunction(getattr(object, attr)) or inspect.ismethod(getattr(object, attr)))
        ]
        py_attr = []

        # Separate attributes into Python and Qobj types
        for attr in attributes:
            value = getattr(object, attr)
            if isinstance(value, Qobj) or (
                isinstance(value, (list, np.ndarray)) and all(isinstance(item, Qobj) for item in value)
            ):
                # Save Qobj attributes to a file in the temporary directory
                fileio.qsave(value, os.path.join(tmp_dir, str(attr)))
            else:
                py_attr.append(attr)

        # Create a dictionary to store the python attributes names and values
        py_data = {
            "__type__": object.__class__.__name__,
            **{attr: getattr(object, attr) for attr in py_attr},
        }

        # Save the python data to a file in the temporary directory
        with open(os.path.join(tmp_dir, "py_data.pkl"), "wb") as f_pkl:
            dill.dump(py_data, f_pkl)

        # Create a zip file to store all the files
        with zipfile.ZipFile(file_name, "w") as zip_file:
            for root, _, files in os.walk(tmp_dir):
                for file in files:
                    file_path = os.path.join(root, file)
                    rel_path = os.path.relpath(file_path, start=tmp_dir)
                    zip_file.write(file_path, rel_path)

    finally:
        # Remove the temporary directory
        shutil.rmtree(tmp_dir)


def load(file_name):
    """
    Loads the attributes of an object from a zip file,
    creates an instance of the object, and sets the attributes.

    Parameters
    ----------
    file_name : str
        Path to the zip file where the attributes are saved.

    Returns
    -------
    object
        The loaded object.
    """
    if not isinstance(file_name, str):
        raise ValueError("file_name must be a string")

    # Extract the zip file to a temporary directory
    tmp_dir = "tmp"
    os.makedirs(tmp_dir, exist_ok=True)

    try:
        with zipfile.ZipFile(file_name, "r") as zip_file:
            zip_file.extractall(tmp_dir)

        # Load the py_data.pkl file
        with open(os.path.join(tmp_dir, "py_data.pkl"), "rb") as f_pkl:
            py_data = dill.load(f_pkl)

        # Get the object name from the py_data
        cls = getattr(quaccatoo, py_data["__type__"])

        # Create an empty instance
        obj = cls.__new__(cls)

        # Load the attributes from the py_data
        for attr, value in py_data.items():
            if attr != "__type__":
                setattr(obj, attr, value)

        # Load the Qobj attributes from the files
        for file in os.listdir(tmp_dir):
            if file != "py_data.pkl":
                attr_name = os.path.splitext(file)[0]
                value = fileio.qload(os.path.join(tmp_dir, attr_name))
                setattr(obj, attr_name, value)

        return obj

    finally:
        # Remove the temporary directory
        shutil.rmtree(tmp_dir)



"""
This module contains the Analysis class and the plot_histogram method.
"""

import matplotlib.pyplot as plt
import numpy as np
from lmfit import Model
from qutip import Bloch, Qobj, fidelity
from scipy.signal import find_peaks
from scipy.stats import linregress, pearsonr

from ..exp_data.exp_data import ExpData
from ..pulsed_sim.pulsed_sim import PulsedSim


class Analysis:
    """
    The Analysis class contains several methods for data Analysis, such as FFT, fitting and plotting.

    Attributes
    ----------
    experiment : PulsedSim or ExpData
        Experiment object to be analyzed containing the results and variable attributes
    FFT_values : tuple
        Tuple with the frequency values and the FFT values
    FFT_peaks : array
        Array with the peaks of the FFT values
    fit_model : lmfit.Model or list(lmfit.Model)
        List with the fitting model for each result
    fit_params : list
        List with the fitted parameters for each result
    fit_cov : list
        List with the covariance of the fitted parameters for each result
    pearson : float
        Pearson correlation coefficient between two experiments
    exp_comparison : PulsedSim or ExpData
        Experiment to be compared with the first one

    Methods
    -------
    compare_with :
        Load a second experiment to compare with the first one
    plot_comparison :
        Plot the results of the experiment and the comparison experiment
    run_FFT :
        Run the real fast fast Fourier transform for the results and variable attributes of the PulsedSimulation object
    get_peaks_FFT :
        Find the peaks of the FFT values calculated by the run_FFT method
    plot_FFT :
        Plot the FFT values calculated by the run_FFT method
    run_fit :
        Run the fit method from lmfit to fit the results of the experiment with a given model
    plot_fit :
        Plot the results of the experiment with the fitted function
    plot_results :
        Plot the results of the experiment
    plot_bloch :
        Plot the results of the experiment in a Bloch sphere if the quantum system has dimension of two
    """

    def __init__(self, experiment):
        """
        Class constructor for Analysis. It takes a PulsedSim or ExpData object as input and checks if the results and variable attributes are not empty and have the same length.

        Parameters
        ----------
        experiment : PulsedSim or ExpData
            Experiment object to be analyzed containing the results and variable attributes
        """
        if isinstance(experiment, (ExpData, PulsedSim)):
            self.experiment = experiment
        else:
            raise ValueError("experiment must be a PulsedSimulation or ExpData object")

        if not isinstance(experiment.results, np.ndarray) and not (
            isinstance(experiment.results, list)
            and all(isinstance(res, np.ndarray) for res in experiment.results)
        ):
            raise ValueError(
                "Results attribute of the experiment must be a numpy array or a list of numpy arrays"
            )

        if len(experiment.results) != len(experiment.variable) and any(
            len(experiment.variable) != len(res) for res in experiment.results
        ):
            raise ValueError("Results and Variable attributes of experiment must have the same length")

        self.FFT_values = []
        self.FFT_peaks = []
        # the fit attributes need to be lists of the same length as the results attribute to avoid index errors
        self.fit_model = [None] * len(self.experiment.results)
        self.fit_params = [None] * len(self.experiment.results)
        self.fit_cov = [None] * len(self.experiment.results)
        self.pearson = None
        self.exp_comparison = None

    def compare_with(self, exp_comparison, results_index=0, comparison_index=0, linear_fit=True):
        """
        Loads a second experiment to compare with the first one.
        If linear_fit is True, a linear fit is performed between the two data sets, which is common for optical experiments.
        Otherwise the Pearson correlation coefficient without linear fit is calculated between the two data sets.

        Parameters
        ----------
        exp_comparison : PulsedSim or ExpData
            Experiment to be compared with the first one
        results_index : int
            Index of the results to be compared if the results attribute is a list
        comparison_index : int
            Index of the results to be compared if the results attribute of the exp_comparison is a list
        linear_fit : bool
            Boolean indicating whether or not to perform a linear fit between the two data sets

        Returns
        -------
        r : float
            Pearson correlation coefficient between the two experiments
        """
        if not isinstance(exp_comparison, PulsedSim) and not isinstance(exp_comparison, ExpData):
            raise ValueError("experiment_2 must be a PulsedSim or ExpData object")

        if not isinstance(results_index, int) or not isinstance(comparison_index, int):
            raise ValueError("results_index and comparison_index must be integers")
        elif (
            results_index > len(self.experiment.results) - 1
            or comparison_index > len(exp_comparison.results) - 1
        ):
            raise ValueError("results_index and comparison_index must be less than the number of results")

        if not isinstance(linear_fit, bool):
            raise ValueError(
                "linear_fit must be a boolean indicating whether or not to perform a linear fit between the two data sets."
            )

        if len(self.experiment.variable) != len(exp_comparison.variable):
            raise ValueError("The variable attributes of the experiments must have the same length")

        self.exp_comparison = exp_comparison

        # if linear_fit is True, perform a linear fit between the two data sets, otherwise calculate the Pearson correlation coefficient
        if linear_fit:
            if isinstance(exp_comparison.results, np.ndarray):
                r = linregress(exp_comparison.results, self.experiment.results)
                self.exp_comparison.results = r[0] * exp_comparison.results + r[1]
            else:
                # if the results are a list, index the results and perform the linear fit
                r = linregress(
                    exp_comparison.results[comparison_index], self.experiment.results[results_index]
                )
                self.exp_comparison.results = r[0] * exp_comparison.results[comparison_index] + r[1]

            self.pearson = r[2]

        else:
            if isinstance(exp_comparison.results, np.ndarray):
                r = pearsonr(exp_comparison.results, self.experiment.results)
            else:
                r = pearsonr(
                    exp_comparison.results[comparison_index], self.experiment.results[results_index]
                )

            self.exp_comparison.results = exp_comparison.results
            self.pearson = r[0]

        self.pearson = r
        return r

    def plot_comparison(self, figsize=(6, 4), xlabel=None, ylabel="Observable", title="Results Comparisons"):
        """
        Plots the results of the experiment and the comparison experiment.

        Parameters
        ----------
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        self.plot_results(figsize, xlabel, ylabel, title)

        if self.pearson is None:
            raise ValueError("You must run the compare_with method before plotting the comparison")

        plt.scatter(
            self.exp_comparison.variable,
            self.exp_comparison.results,
            label="Compared Experiment",
            alpha=0.7,
            s=15,
        )
        plt.legend(loc="upper right", bbox_to_anchor=(1.2, 1))

    ######################################################## FFT Methods ########################################################

    def run_FFT(self):
        """
        Run the real fast Fourier transform for the results and variable attributes of the PulsedSimulation object.
        The results are centered around the mean value before the FFT is calculated in order to remove the DC component.

        Returns
        -------
        FFT_values : tuple
            Tuple with the frequency values and the FFT values
        """
        if isinstance(self.experiment.results, np.ndarray):
            y = np.abs(np.fft.rfft(self.experiment.results - np.mean(self.experiment.results)))

        elif isinstance(self.experiment.results, list) and all(
            isinstance(res, np.ndarray) for res in self.experiment.results
        ):
            y = [np.abs(np.fft.rfft(res - np.mean(res))) for res in self.experiment.results]

        else:
            raise ValueError("Results must be a numpy array or a list of numpy arrays")

        freqs = np.fft.rfftfreq(
            len(self.experiment.variable), self.experiment.variable[1] - self.experiment.variable[0]
        )

        self.FFT_values = (freqs, y)

        return self.FFT_values

    def get_peaks_FFT(self, **find_peaks_kwargs):
        """
        Find the peaks of the FFT values calculated by the run_FFT method.

        Parameters
        ----------
        find_peaks_kwargs : dict
            Dictionary with the arguments to be passed to the scipy.signal.find_peaks function

        Returns
        -------
        FFT_peaks : array
            Array with the peaks of the FFT values
        """
        if len(self.FFT_values) == 0:
            raise ValueError("No FFT values to analyze, you must run the FFT first")

        if isinstance(self.experiment.results, np.ndarray):
            self.FFT_peaks_index = find_peaks(self.FFT_values[1], **find_peaks_kwargs)
            self.FFT_peaks = self.FFT_values[0][self.FFT_peaks_index[0]]

        elif isinstance(self.experiment.results, list):
            self.FFT_peaks_index = [find_peaks(FFT, **find_peaks_kwargs) for FFT in self.FFT_values[1]]
            self.FFT_peaks = [self.FFT_values[0][index[0]] for index in self.FFT_peaks_index]
        else:
            raise ValueError("Results must be a numpy array or a list of numpy arrays")

        return self.FFT_peaks

    def plot_FFT(
        self, freq_lim=None, figsize=(6, 4), xlabel=None, ylabel="FFT Intensity", title="FFT of the Results"
    ):
        """
        Plots the FFT

        Parameters
        ----------
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        if len(self.FFT_values) == 0:
            raise ValueError("No FFT values to plot, you must run the FFT first")

        # check if figsize is a tuple of two positive floats
        if not (isinstance(figsize, tuple) or len(figsize) == 2):
            raise ValueError("figsize must be a tuple of two positive floats")

        if xlabel is None and isinstance(self.experiment, PulsedSim):
            xlabel = f"Frequency ({self.experiment.system.units_H0})"
        elif xlabel is None and isinstance(self.experiment, ExpData):
            xlabel = "Frequency"
        elif not isinstance(xlabel, str):
            raise ValueError("xlabel must be a string")

        # initialize the figure and axis for the plot
        fig, ax = plt.subplots(1, 1, figsize=figsize)

        # if the FFT_values[1] is an array plot it, otherwise if it is a list iterate over the elements and plot each one
        if isinstance(self.FFT_values[1], np.ndarray):
            ax.plot(self.FFT_values[0], self.FFT_values[1])
            if len(self.FFT_peaks) != 0:
                ax.scatter(
                    self.FFT_peaks,
                    self.FFT_values[1][self.FFT_peaks_index[0]],
                    color="red",
                    label="Peaks",
                    s=50,
                )

        elif isinstance(self.FFT_values[1], list):
            # if the FFT_peaks attribute is not empty, then plot them with the FFT
            if len(self.FFT_peaks) != 0:
                for itr in range(len(self.FFT_values[1])):
                    ax.plot(self.FFT_values[0], self.FFT_values[1][itr], label=f"FFT {itr}")
                    ax.scatter(
                        self.FFT_peaks[itr],
                        self.FFT_values[1][itr][self.FFT_peaks_index[itr][0]],
                        color="red",
                        label=f"Peaks {itr}",
                        s=50,
                    )
            else:
                for itr in range(len(self.FFT_values[1])):
                    ax.plot(self.FFT_values[0], self.FFT_values[1][itr], label=f"FFT {itr}")

        # set the x-axis limits to the total time of the experiment
        if freq_lim is None:
            ax.set_xlim(self.FFT_values[0][0], self.FFT_values[0][-1])
        elif len(freq_lim) == 2:
            ax.set_xlim(freq_lim[0], freq_lim[1])
        else:
            raise ValueError("freq_lim must be a tuple of two floats")

        # set the axes labels according to the parameters
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)

    ######################################################## FIT Methods ########################################################

    def run_fit(self, fit_model, results_index=0, guess=None):
        """
        Run the fit method from lmfit to fit the results of the experiment with a given model,
        guess for the initial parameters.

        Parameters
        ----------
        fit_model : lmfit.Model
            Model to be used to fit the results
        results_index : int
            Index of the results to be fitted if the results attribute is a list
        guess : dict
            Initial guess for the parameters of the model
            Takes a dictionary consisting of parameter names as the keys and their initial guess as the value.
            See the definitions of the models in the fit_functions.py file for details.

        Examples
        --------
        The typical usage scheme is to pass a model and optionally, a dictionary of initial values. The file
        fit_functions.py includes a bunch of predefined models and imports some commonly needed ones from lmfit
        which can be directly passed to the function and used. Alternatively, the Model function from lmfit takes
        a custom Python function and instantiates a model class object with it, which can then be used here.

        >>> from lmfit import Model #needed for custom models
        >>> # my_analysis_obj is an instance of the analysis class
        >>> my_analysis_obj.run_fit(fit_model=RabiModel())
        >>> my_analysis_obj.run_fit(
                fit_model=Model(fit_two_lorentz_sym),
                guess = {'A': 0.5, 'gamma': 0.2, 'f_mean':1749, 'f_delta':3,'C':1}
                )

        In the snippet above, we showcase some ways to perform fits using this method. Using predefined
        models takes care of most of the stuff. One can still pass a guess dictionary to provide better
        initial values. The second example shows the usage with a custom model instantiated from a function
        along with user provided guesses. It's important to note that the first parameter of these custom
        functions should be `x`. Moreover, the keys in the dictionary should correspond to the other
        parameters of the function (this holds true for the predefined models as well). The names of the
        parameters can be referred to from the source file (fit_functions.py).

        Returns
        -------
        fit_params : dict
            best fit parameter values of parameters as a dict with the parameter names as the keys

        """
        if not isinstance(fit_model, Model):
            raise TypeError(
                "fit_model must be an instance of lmfit.Model. Remember to instantiate the class by adding parentheses."
            )

        # if there is only one result, just fit the results with the model
        if isinstance(self.experiment.results, np.ndarray):
            self.fit_model = fit_model
            if guess:
                self.fit_params = fit_model.fit(self.experiment.results, x=self.experiment.variable, **guess)
            else:
                try:
                    params = fit_model.guess(self.experiment.results, x=self.experiment.variable)
                    self.fit_params = fit_model.fit(
                        self.experiment.results, x=self.experiment.variable, params=params
                    )
                except NotImplementedError:
                    params = fit_model.make_params()
                    self.fit_params = fit_model.fit(
                        self.experiment.results, x=self.experiment.variable, params=params
                    )

            return self.fit_params.best_values

        # if there are multiple results, check if the results_index is an integer and if it is less than the number of results then fit
        elif isinstance(self.experiment.results, list):
            if (
                not isinstance(results_index, int)
                or results_index < 0
                or results_index >= len(self.experiment.results)
            ):
                raise ValueError(
                    "results_index must be a non-negative integer less than the number of results"
                )

            self.fit_model[results_index] = fit_model
            if guess:
                self.fit_params[results_index] = fit_model.fit(
                    self.experiment.results[results_index], x=self.experiment.variable, **guess
                )
            else:
                try:
                    params = fit_model.guess(
                        self.experiment.results[results_index], x=self.experiment.variable
                    )
                    self.fit_params[results_index] = fit_model.fit(
                        self.experiment.results[results_index], x=self.experiment.variable, params=params
                    )
                except NotImplementedError:
                    params = fit_model.make_params()
                    self.fit_params[results_index] = fit_model.fit(
                        self.experiment.results[results_index], x=self.experiment.variable, params=params
                    )
            return self.fit_params[results_index].best_values

    def plot_fit(self, figsize=(6, 4), xlabel=None, ylabel="Expectation Value", title="Pulsed Result"):
        """
        Plot the results of the experiment with the fitted function.

        Parameters
        ----------
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        self.plot_results(figsize, xlabel, ylabel, title)

        if isinstance(self.experiment.results, np.ndarray):
            plt.plot(self.experiment.variable, self.fit_params.best_fit, label="Fit")

        elif isinstance(self.experiment.results, list):
            for itr in range(len(self.experiment.results)):
                if self.fit_model[itr] is not None:
                    plt.plot(self.experiment.variable, self.fit_params[itr].best_fit, label=f"Fit {itr}")

        plt.legend(loc="upper right", bbox_to_anchor=(1.2, 1))

    ######################################################## Other Plotting Methods ########################################################

    def plot_results(self, figsize=(6, 4), xlabel=None, ylabel="Observable", title="Results"):
        """
        Plots the results of the experiment

        Parameters
        ----------
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        if not (isinstance(figsize, tuple) or len(figsize) == 2):
            raise ValueError("figsize must be a tuple of two positive floats")

        if not isinstance(ylabel, str) or not isinstance(title, str):
            raise ValueError("ylabel and title must be strings")

        if xlabel is None:
            xlabel = self.experiment.variable_name
        elif not isinstance(xlabel, str):
            raise ValueError("xlabel must be a string")

        fig, ax = plt.subplots(1, 1, figsize=figsize)

        # check if the observable is a Qobj or a list of Qobj
        if isinstance(self.experiment.results, np.ndarray):
            ax.plot(self.experiment.variable, self.experiment.results, lw=2, alpha=0.7, label="Observable")

        elif isinstance(self.experiment.results, list):
            # if it is a list, iterate over the observables and plot each one
            for itr in range(len(self.experiment.system.observable)):
                # plot all observables in the results
                ax.plot(
                    self.experiment.variable,
                    self.experiment.results[itr],
                    label=f"Observable {itr}",
                    lw=2,
                    alpha=0.7,
                )

        else:
            raise ValueError("Results must be a numpy array or a list of numpy arrays")

        # set the x-axis limits to the variable of the experiment
        ax.set_xlim(self.experiment.variable[0], self.experiment.variable[-1])

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.legend(loc="upper right", bbox_to_anchor=(1.2, 1))
        ax.set_title(title)

    def plot_bloch(self, figsize=(6, 4)):
        """
        Plot the results of the experiment in a Bloch sphere if the quantum system has dimension of two.

        Parameters
        ----------
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        """
        if not isinstance(self.experiment, PulsedSim):
            raise ValueError("experiment must be a PulsedSim object")

        if len(self.experiment.rho) == 1:
            raise ValueError("Density matrices were not calculated, please run experiment first.")
        elif isinstance(self.experiment.rho, list) and all(rho.shape[0] == 2 for rho in self.experiment.rho):
            pass
        else:
            raise ValueError("QSys must have dimension of two to be able to plot a Bloch sphere")

        if not (isinstance(figsize, tuple) or len(figsize) == 2):
            raise ValueError("figsize must be a tuple of two positive floats")

        fig, axs = plt.subplots(1, 1, figsize=figsize, subplot_kw={"projection": "3d"})

        colors = plt.cm.viridis(np.linspace(0, 1, len(self.experiment.rho)))

        bloch = Bloch(fig)
        bloch.add_states(self.experiment.rho, kind="point", colors=colors)
        bloch.frame_alpha = 0
        bloch.render()


def plot_histogram(rho, rho_comparison=None, component="real", figsize=(5, 5), title="Matrix Histogram"):
    """
    Plot a 3D histogram of the final density matrix of the simulation.

    Parameters
    ----------
    component : str
        Component of the density matrix to be plotted. Can be 'real', 'imag', or 'abs'.
    figsize : tuple
        Size of the figure to be passed to matplotlib.pyplot
    rho_comparison : Qobj, optional
        A Qobj representing a density matrix to be compared with the final density matrix of the simulation
    title : str
        Title of the plot
    """
    # Check all parameters
    if component not in ["real", "imag", "abs"]:
        raise ValueError("component must be 'real', 'imag', or 'abs'")

    if not (isinstance(figsize, tuple) or len(figsize) == 2):
        raise ValueError("figsize must be a tuple of two positive floats")

    if isinstance(rho, Qobj) and rho.shape[0] == rho.shape[1]:
        N = rho.shape[0]
        rho = rho.full()
    else:
        raise ValueError("rho must be a Qobj representing a square density matrix")

    if rho_comparison is None:
        pass
    elif isinstance(rho_comparison, Qobj) and rho_comparison.shape[0] == rho_comparison.shape[1] == N:
        rho_comparison = rho_comparison.full()
    else:
        raise ValueError(
            "rho_comparison must be a Qobj with the same shape as the density matrix to be compared with"
        )

    # Create 3D plot
    fig = plt.figure(figsize=(3, 3))
    ax = fig.add_subplot(111, projection="3d")
    ax.view_init(azim=-50, elev=30)

    # Dimensions and positions for the bars
    xpos, ypos = np.meshgrid(np.arange(N), np.arange(N))
    xpos, ypos = xpos.flatten(), ypos.flatten()
    zpos = np.zeros_like(xpos)
    dx = dy = 0.5 * np.ones_like(zpos)

    # Prepare the data for the bars depending on the component
    if component == "real":
        dz_sim = np.real(rho).flatten()
        color = "C0"
        if rho_comparison is not None:
            dz_comp = np.real(rho_comparison).flatten()

    elif component == "imag":
        dz_sim = np.imag(rho).flatten()
        color = "C1"
        if rho_comparison is not None:
            dz_comp = np.imag(rho_comparison).flatten()

    elif component == "abs":
        dz_sim = np.abs(rho).flatten()
        color = "C2"
        if rho_comparison is not None:
            dz_comp = np.abs(rho_comparison).flatten()

    # Plot the bars
    ax.bar3d(xpos, ypos, zpos, dx, dy, dz_sim, color=color, linewidth=0, alpha=0.3, zorder=0)

    if rho_comparison is not None:
        print(f"Fidelity: {fidelity(Qobj(rho), Qobj(rho_comparison))}")
        ax.bar3d(
            xpos,
            ypos,
            zpos,
            dx,
            dy,
            dz_comp,
            edgecolor="k",
            linewidth=1,
            linestyle="dotted",
            alpha=0,
            zorder=1,
        )

    # Aesthetics, labels, and title
    ax.grid(False)
    ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    fig.suptitle(title, fontsize=12, x=0.2, y=0.85)
    ax.set_xticks(np.arange(N) + 0.5)
    ax.set_yticks(np.arange(N) + 0.5)
    ax.set_xticklabels([f"$|{i}\\rangle$" for i in range(N)])
    ax.set_yticklabels([f"$|{i}\\rangle$" for i in range(N)])
    ax.set_zticks([-1, -0.5, 0, 0.5, 1.0])



import numpy as np
import pytest
from lmfit import Model
from qutip import basis, fock_dm, sigmax, sigmay, sigmaz

from quaccatoo import NV, PMR, XY8, Analysis, ExpData, Hahn, QSys, Rabi, square_pulse
from quaccatoo.analysis.fit_functions import (
    ExpDecayModel,
    GaussianModel,
    RabiModel,
    fit_two_lorentz_sym,
)

"""
The testing framework for QuaCCAToo.

We use `pytest` for testing. See https://docs.pytest.org/en/stable/getting-started.html
for a quick introduction to pytest.

The gist is to use fixtures (`@pytest.fixture` decorator) for objects which are needed
repeatedly, and group related tests into an appropriately named test class.

Remember to mark long running tests with the `@pytest.mark.slow` decorator. These can then be run
with the `--runslow` CLI flag passed to `pytest`.
"""


# QSys fixture for reuse in multiple tests
@pytest.fixture
def qsys():
    delta = 1
    return QSys(H0=delta / 2 * sigmaz(), rho0=fock_dm(2, 0), observable=sigmaz(), units_H0="MHz")


# Test if the eigenstates of the qsys fixture are correct
class TestQSys:
    def test_states(self, qsys):
        assert (qsys.eigenstates[0], qsys.eigenstates[1]) == (
            -basis(2, 1),
            -basis(2, 0),
        )

    def test_levels(self, qsys):
        assert np.array_equal(qsys.energy_levels, np.array([0, 1]))


# Rabi object (fixture) used in the TestRabi class below
@pytest.fixture
def rabi_exp(qsys):
    w1 = 0.1
    delta = 1

    def custom_pulseX(t):
        return np.cos(delta * t)

    def custom_pulseY(t):
        return np.cos(delta * t - np.pi / 2)

    return Rabi(
        pulse_duration=np.linspace(0, 40, 1000),
        system=qsys,
        H1=[w1 * sigmax() / 2, w1 * sigmay() / 2],
        pulse_shape=[custom_pulseX, custom_pulseY],
    )


class TestRabi:
    # Uses the rabi fixture defined above to check if the rabi frequency fit
    # is close to the expected value
    def test_tpi(self, rabi_exp):
        rabi_exp.run()
        rabi_analysis = Analysis(rabi_exp)
        rabi_analysis.run_fit(fit_model=RabiModel())
        assert np.isclose(rabi_analysis.fit_params.best_values["Tpi"], 5, atol=1e-3)


# Hahn object (fixture) used in the Test class below
@pytest.fixture
def hahn_exp(qsys):
    w1 = 0.1
    delta = 1
    gamma = 0.1

    qsys.c_ops = gamma * sigmaz()

    return Hahn(
        free_duration=np.linspace(5, 25, 30),
        pi_pulse_duration=1 / 2 / w1,
        projection_pulse=True,
        system=qsys,
        H1=w1 * sigmax(),
        pulse_shape=square_pulse,
        pulse_params={"f_pulse": delta},
    )


class TestHahn:
    # @pytest.mark.slow
    @pytest.mark.filterwarnings("ignore::DeprecationWarning")

    # Uses the Hahn fixture defined above to check if the decay rate
    # is close to the expected value
    def test_decay(self, hahn_exp):
        hahn_exp.run()
        hahn_analysis = Analysis(hahn_exp)
        hahn_analysis.run_fit(fit_model=ExpDecayModel())
        assert np.isclose(hahn_analysis.fit_params.best_values["Tc"], 3.953, atol=1e-3)


class TestXY8:
    @pytest.mark.slow

    # Runs the XY8 sequence on an NV object
    # and checks if the center of the peak is in the expected position.
    # We don't use an outside fixture here since we need handcrafted values for this test
    def test_xy8_exp(self):
        qsys = NV(
            N=15,
            B0=39.4,
            units_B0="mT",
            theta=2.6,
            units_angles="deg",
        )
        w1 = 20
        XY8_15N = XY8(
            M=2,
            free_duration=np.linspace(0.25, 0.36, 100),
            pi_pulse_duration=1 / 2 / w1,
            system=qsys,
            H1=w1 * qsys.MW_H1,
            pulse_params={"f_pulse": qsys.MW_freqs[1]},
            time_steps=100,
        )
        XY8_15N.run()
        XY8_analysis = Analysis(XY8_15N)
        XY8_analysis.run_fit(fit_model=GaussianModel())
        assert 0.29 <= XY8_analysis.fit_params.best_values["center"] <= 0.31


class TestPODMR:
    @pytest.mark.slow

    # Runs the PODMR sequence on an NV object
    # and checks if the frequencies are close to the expected values
    # We don't use an outside fixture here since we need handcrafted values for this test
    def test_podmr(self):
        qsys = NV(N=15, B0=40, units_B0="mT")
        w1 = 0.3

        podmr_exp = PMR(
            frequencies=np.arange(1745, 1753, 0.05),
            pulse_duration=1 / 2 / w1,
            system=qsys,
            H1=w1 * qsys.MW_H1,
        )

        podmr_exp.run()
        podmr_analysis = Analysis(podmr_exp)

        podmr_analysis.run_fit(
            fit_model=Model(fit_two_lorentz_sym),
            guess={"A": 0.5, "gamma": 0.2, "f_mean": 1749, "f_delta": 3, "C": 1},
        )
        assert np.isclose(
            podmr_analysis.fit_params.best_values["f_mean"], 1.749e3, atol=1e-3
        ) and np.isclose(podmr_analysis.fit_params.best_values["f_delta"], 3.029, atol=1e-3)


class TestExpData:
    def test_expdata(self):
        qsys_exp = NV(
            N=15,
            units_B0="mT",
            B0=38.4,
        )
        exp_data = ExpData(file_path="./tests/data//Ex02_NV_rabi.dat")
        w1_exp = 16.72

        rabi_sim_exp = Rabi(
            pulse_duration=np.arange(0, 0.15, 3e-3),
            system=qsys_exp,
            H1=w1_exp * qsys_exp.MW_H1,
            pulse_params={"f_pulse": qsys_exp.MW_freqs[0]},
        )

        rabi_sim_exp.run()
        rabi_analysis_exp = Analysis(rabi_sim_exp)
        exp_data.variable *= 1e6
        rabi_analysis_exp.compare_with(exp_data)
        rabi_analysis_exp.pearson
        print(rabi_analysis_exp.pearson.intercept)
        assert np.isclose(rabi_analysis_exp.pearson.slope, 2.5895, atol=1e-3) and np.isclose(
            rabi_analysis_exp.pearson.intercept, -2.0453, atol=1e-3
        )



"""
This module contains dynamical decoupling pulse sequences, used in quantum sensing and for extending coherence of quantum systems.
"""

import numpy as np
from qutip import Qobj

from .pulse_shapes import square_pulse
from .pulsed_sim import PulsedSim

####################################################################################################


class CPMG(PulsedSim):
    """
    This class contains a Carr-Purcell-Meiboom-Gill sequence used in quantum sensing experiments.

    The CPMG sequence consists of a series of pi pulses and free evolution times,
    such that these periodicals inversions will cancel out oscillating noises except for frequencies corresponding to the pulse separation.

    Methods
    -------
    CPMG_sequence :
        defines the Carr-Purcell-Meiboom-Gill sequence for a given free evolution time tau and the set of attributes defined in the constructor,
        returning the final state. The sequence is to be called by the parallel_map method of QuTip.
    _get_pulse_profiles :
        Generates the pulse profiles for the CPMG sequence for a given tau. The pulse profiles are stored in the pulse_profiles attribute of the object.
    plot_pulses :
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the CPMG sequence for a given tau and then plot them.

    Notes
    -----
    The CPMG sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self,
        free_duration,
        system,
        M,
        pi_pulse_duration,
        H1,
        H2=None,
        projection_pulse=True,
        pulse_shape=square_pulse,
        pulse_params=None,
        options=None,
        time_steps=100,
    ):
        """
        Class constructor for the Carr-Purcell-Meiboom-Gill sequence

        Parameters
        ----------
        free_duration : numpy array
            Time array for the simulation representing the free evolution time to be used as the variable attribute for the simulation
        system : QSys
            Quantum system object containing the initial state, internal time independent Hamiltonian and collapse operators
        M : int
            Order of the CPMG sequence
        H1 : Qobj or list(Qobj)
            Control Hamiltonian of the system
        pi_pulse_duration : float or int
            Duration of the pi pulse
        H2 : list(Qobj or function)
            Time dependent sensing Hamiltonian of the system
        projection_pulse : bool
            Boolean to determine if a final pi/2 pulse is to be included in order to project the measurement into the Sz basis
        pulse_shape : FunctionType or list(FunctionType)
            Pulse shape function or list of pulse shape functions representing the time modulation of H1
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions
        time_steps : int
            Number of time steps in the pulses for the simulation
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1,
            pulse_shape,
            pulse_params,
            options,
            time_steps,
            free_duration,
            pi_pulse_duration,
            M,
            projection_pulse,
        )
        self.sequence = self.CPMG_sequence

        base_pulse = pulse_params.copy()
        self.pulse_params = [{**base_pulse, "phi_t": 0}, {**base_pulse, "phi_t": -np.pi / 2}]

    def CPMG_sequence(self, tau):
        """
        Defines the CPMG sequence for a given free evolution time tau and the set of attributes defined in the constructor.
        The sequence consists of an initial pi/2 pulse, and M pi-pulses separated by free evolution time tau.
        If projection_pulse is True, the sequence will include a final pi/2 pulse on Y axis to project the measurement into the Sz basis.
        The sequence is to be called by the parallel_map method of QuTip.

        Parameters
        ----------
        tau : float
            Free evolution time

        Returns
        -------
        rho : Qobj
            Final state
        """
        ps = tau - self.pi_pulse_duration

        # initial pi/2 pulse on Y
        self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params[1])
        self._free_evolution(ps / 2 - self.pi_pulse_duration / 2, self.options)

        # repeat M-1 times the pi pulse and free evolution of ps
        for itr_M in range(self.M - 1):
            self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params[0])
            self._free_evolution(ps, self.options)

        # perform the last pi pulse on X
        self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params[0])

        if self.projection_pulse:
            self._free_evolution(ps / 2 - self.pi_pulse_duration / 2, self.options)
            # final pi/2 pulse on Y
            self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params[1])
        else:
            self._free_evolution(ps / 2, self.options)

        return self.rho

    def _get_pulse_profiles(self, tau=None):
        """
        Generates the pulse profiles for the CPMG sequence for a given tau.

        Parameters
        ----------
        tau : float
            Free evolution variable or pulse spacing for the  sequence.
        """
        # check if tau is correctly defined
        if tau is None:
            tau = self.variable[-1]
        elif not isinstance(tau, (int, float)) or tau < self.pi_pulse_duration:
            raise ValueError("tau must be a positive real number larger than pi_pulse_duration")

        self.pulse_profiles = []

        # add the first pi/2 pulse on Y
        if isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params[1],
                ]
            )
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params[1],
                    ]
                    for i in range(len(self.H1))
                ]
            )

        t0 = self.pi_pulse_duration / 2
        ps = tau - self.pi_pulse_duration

        # add the first free evolution
        self.pulse_profiles.append([None, [t0, t0 + ps / 2 - self.pi_pulse_duration / 2], None, None])
        t0 += ps / 2 - self.pi_pulse_duration / 2

        # add pulses and free evolution 8*M-1 times
        for itr_M in range(8 * self.M - 1):
            # add a pi pulse on X
            if isinstance(self.H1, Qobj):
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params[0],
                    ]
                )
            elif isinstance(self.H1, list):
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params[0],
                        ]
                        for i in range(len(self.H1))
                    ]
                )
            t0 += self.pi_pulse_duration
            # add a free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps], None, None])
            t0 += ps

        # add another pi pulse on X
        if isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params[0],
                ]
            )
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params[0],
                    ]
                    for i in range(len(self.H1))
                ]
            )
        t0 += self.pi_pulse_duration

        if self.projection_pulse:
            # add the last free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps / 2 - self.pi_pulse_duration / 2], None, None])
            t0 += ps / 2 - self.pi_pulse_duration / 2

            if isinstance(self.H1, Qobj):
                # add the last pi/2 pulse
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params[1],
                    ]
                )
            elif isinstance(self.H1, list):
                # add the first pi/2 pulse
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params[1],
                        ]
                        for i in range(len(self.H1))
                    ]
                )

            t0 += self.pi_pulse_duration / 2
        else:
            # add the last free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps / 2], None, None])
            t0 += ps / 2

        self.total_time = t0

    def plot_pulses(
        self, figsize=(6, 6), xlabel=None, ylabel="Expectation Value", title="Pulse Profiles", tau=None
    ):
        """
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the CPMG sequence for a given tau and then plot them.

        Parameters
        ----------
        tau : float
            Free evolution time for the  sequence. Contrary to the run method, the free evolution must be a single number in order to plot the pulse profiles.
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        self._get_pulse_profiles(tau)
        # call the plot_pulses method of the parent class
        super().plot_pulses(figsize, xlabel, ylabel, title)


####################################################################################################


class XY(PulsedSim):
    """
    This class contains the XY-M pulse sequence.

    The sequence is composed of intercalated X and Y pi pulses and free evolutions repeated M times.
    It acts similar to the CPMG sequence, but the alternation of the pulse improves noise suppression on different axis.

    Methods
    -------
    XY_sequence :
        Defines the XY sequence for a given free evolution time tau and the set of attributes defined in the constructor,
        returning the final state. The sequence is to be called by the parallel_map method of QuTip.
    _get_pulse_profiles :
        Generates the pulse profiles for the XY-M sequence for a given tau. The pulse profiles are stored in the pulse_profiles attribute of the object.
    plot_pulses :
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the XY-M sequence for a given tau and then plot them.

    Notes
    -----
    The XY sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self,
        free_duration,
        system,
        M,
        pi_pulse_duration,
        H1,
        H2=None,
        projection_pulse=True,
        pulse_shape=square_pulse,
        pulse_params=None,
        options=None,
        time_steps=100,
    ):
        """
        Class constructor for the XY sequence

        Parameters
        ----------
        free_duration : numpy array
            Time array for the simulation representing the free evolution time to be used as the variable attribute for the simulation
        system : QSys
            Quantum system object containing the initial state, internal time independent Hamiltonian and collapse operators
        M : int
            Order of the XY sequence
        H1 : Qobj, list(Qobj)
            Control Hamiltonian of the system
        pi_pulse_duration : float, int
            Duration of the pi pulse
        H2 : Qobj, list(Qobj), optional
            Time dependent sensing Hamiltonian of the system
        pulse_shape : FunctionType, list(FunctionType), optional
            Pulse shape function or list of pulse shape functions representing the time modulation of H1
        pulse_params : dict, optional
            Dictionary of parameters for the pulse_shape functions
        time_steps : int, optional
            Number of time steps in the pulses for the simulation
        options : dict, optional
            Dictionary of solver options
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1,
            pulse_shape,
            pulse_params,
            options,
            time_steps,
            free_duration,
            pi_pulse_duration,
            M,
            projection_pulse,
        )
        self.sequence = self.XY_sequence

        base_pulse = pulse_params.copy()
        self.pulse_params = [{**base_pulse, "phi_t": 0}, {**base_pulse, "phi_t": -np.pi / 2}]

    def XY_sequence(self, tau):
        """
        Defines the XY-M composed of intercalated pi pulses on X and Y axis with free evolutions of time tau repeated M times.
        If projection_pulse is True, the sequence will include a final pi/2 pulse on X axis to project the measurement into the Sz basis.
        The sequence is to be called by the parallel_map method of QuTip.

        Parameters
        ----------
        tau : float
            Free evolution time

        Returns
        -------
        rho : Qobj
            Final state
        """
        ps = tau - self.pi_pulse_duration

        # initial pi/2 pulse on X
        self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params[0])
        self._free_evolution(ps / 2 - self.pi_pulse_duration / 2, self.options)

        # repeat M-1 times the pi X pulse, free evolution of ps, pi Y pulse and free evolution of ps
        for itr_M in range(2 * self.M - 1):
            self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params[itr_M % 2])
            self._free_evolution(ps, self.options)

        self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params[1])

        if self.projection_pulse:
            self._free_evolution(ps / 2 - self.pi_pulse_duration / 2, self.options)
            # final pi/2 pulse on X
            self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params[0])
        else:
            self._free_evolution(ps / 2, self.options)

        return self.rho

    def _get_pulse_profiles(self, tau=None):
        """
        Generates the pulse profiles for the XY-M sequence for a given tau.
        The pulse profiles are stored in the pulse_profiles attribute of the object.

        Parameters
        ----------
        tau : float
            free evolution variable or pulse spacing for the  sequence
        """
        # checl if tau is correctly define
        if tau is None:
            tau = self.variable[-1]
        elif not isinstance(tau, (int, float)) or tau < self.pi_pulse_duration:
            raise ValueError("tau must be a positive real number larger than pi_pulse_duration")

        self.pulse_profiles = []

        # add the first pi/2 pulse on X axis
        if isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params[0],
                ]
            )
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params[0],
                    ]
                    for i in range(len(self.H1))
                ]
            )

        t0 = self.pi_pulse_duration / 2
        ps = tau - self.pi_pulse_duration

        # add the first free evolution
        self.pulse_profiles.append([None, [t0, t0 + ps / 2 - self.pi_pulse_duration / 2], None, None])
        t0 += ps / 2 - self.pi_pulse_duration / 2

        # add pulses and free evolution M-1 times
        for itr_M in range(2 * self.M - 1):
            # add a pi pulse
            if isinstance(self.H1, Qobj):
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params[itr_M % 2],
                    ]
                )
            elif isinstance(self.H1, list):
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params[itr_M % 2],
                        ]
                        for i in range(len(self.H1))
                    ]
                )
            t0 += self.pi_pulse_duration

            # add a free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps], None, None])
            t0 += ps

        # add another pi pulse
        if isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params[1],
                ]
            )
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params[1],
                    ]
                    for i in range(len(self.H1))
                ]
            )
        t0 += self.pi_pulse_duration

        if self.projection_pulse:
            # add the last free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps / 2 - self.pi_pulse_duration / 2], None, None])
            t0 += ps / 2 - self.pi_pulse_duration / 2

            if isinstance(self.H1, Qobj):
                # add the last pi/2 pulse
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params[0],
                    ]
                )
            elif isinstance(self.H1, list):
                # add the first pi/2 pulse
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params[0],
                        ]
                        for i in range(len(self.H1))
                    ]
                )

            t0 += self.pi_pulse_duration / 2
        else:
            # add the last free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps / 2], None, None])
            t0 += ps / 2

        self.total_time = t0

    def plot_pulses(
        self, figsize=(6, 6), xlabel=None, ylabel="Expectation Value", title="Pulse Profiles", tau=None
    ):
        """
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the XY-M sequence for a given tau and then plot them.

        Parameters
        ----------
        tau : float
            Free evolution time for the  sequence. Contrary to the run method, the free evolution must be a single number in order to plot the pulse profiles.
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        # generate the pulse profiles for the given tau
        self._get_pulse_profiles(tau)

        # call the plot_pulses method of the parent class
        super().plot_pulses(figsize, xlabel, ylabel, title)


####################################################################################################


class XY8(PulsedSim):
    """
    This contains the XY8-M sequence.

    The XY8-M is a further improvement from the XY-M sequence, where the X and Y pulses are group antisymmetrically in pairs of 4 as X-Y-X-Y-Y-X-Y-X,
    in order to improve noise suppression and pulse errors.

    Methods
    -------
    XY_sequence :
        Defines the XY8-M sequence for a given free evolution time tau and the set of attributes defined in the constructor,
        returning the final state. The sequence is to be called by the parallel_map method of QuTip.
    _get_pulse_profiles :
        Generates the pulse profiles for the XY8-M sequence for a given tau. The pulse profiles are stored in the pulse_profiles attribute of the object.
    plot_pulses :
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the XY8-M sequence for a given tau and then plot them.

    Notes
    -----
    The XY8 sequence inherits the methods and attributes from the PulsedSim class.
    """

    def __init__(
        self,
        free_duration,
        system,
        M,
        pi_pulse_duration,
        H1,
        H2=None,
        projection_pulse=True,
        pulse_shape=square_pulse,
        pulse_params=None,
        options=None,
        time_steps=100,
        RXY8=False,
    ):
        """
        Class constructor for the XY8 sequence

        Parameters
        ----------
        free_duration : numpy array
            Time array for the simulation representing the free evolution time to be used as the variable attribute for the simulation
        system : QSys
            Quantum system object containing the initial state, internal time independent Hamiltonian and collapse operators
        M : int
            Order of the XY sequence
        H1 : Qobj, list(Qobj)
            Control Hamiltonian of the system
        pi_pulse_duration : float, int
            Duration of the pi pulse
        H2 : Qobj, list(Qobj), optional
            Time dependent sensing Hamiltonian of the system
        pulse_shape : FunctionType, list(FunctionType), optional
            Pulse shape function or list of pulse shape functions representing the time modulation of H1
        pulse_params : dict, optional
            Dictionary of parameters for the pulse_shape functions
        time_steps : int, optional
            Number of time steps in the pulses for the simulation
        options : dict, optional
            Dictionary of solver options from Qutip
        projection_pulse : bool
            Boolean to determine if a final pi/2 pulse is to be included in order to project the measurement into the Sz basis
        RXY8 : bool
            Boolen to determine if a random phase is to be added to each XY8 block
        """
        super().__init__(system, H2)
        self._check_attr_predef_seqs(
            H1,
            pulse_shape,
            pulse_params,
            options,
            time_steps,
            free_duration,
            pi_pulse_duration,
            M,
            projection_pulse,
        )
        self.sequence = self.XY8_sequence

        if RXY8:
            random_phases = np.random.rand(M) * 2 * np.pi
        elif not RXY8:
            random_phases = np.zeros(M)
        else:
            raise ValueError(
                "RXY8 must be a boolean value indicating weather to add a random phase to each XY8 block or not."
            )

        base_pulse = self.pulse_params.copy()

        self.pulse_params = []
        for itr_M in range(M):
            phi_x = 0 + random_phases[itr_M]
            phi_y = -np.pi / 2 + random_phases[itr_M]

            px = base_pulse.copy()
            py = base_pulse.copy()
            px["phi_t"] = phi_x
            py["phi_t"] = phi_y

            self.pulse_params.extend([px, py, px, py, py, px, py, px])

    def XY8_sequence(self, tau):
        """
        Defines the XY8-M composed of 8 intercalated pi pulses on X and Y axis with free evolutions of time tau repeated M times.
        If random_phase is set to True, a random phase is added in each XY8 block.
        If projection_pulse is True, the sequence will include a final pi/2 pulse on X axis to project the measurement into the Sz basis.
        The sequence is to be called by the parallel_map method of QuTip.

        Parameters
        ----------
        tau : float
            Free evolution time

        Returns
        -------
        rho : Qobj
            Final state
        """
        ps = tau - self.pi_pulse_duration

        # initial pi/2 pulse on X
        self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params[0])
        self._free_evolution(ps / 2 - self.pi_pulse_duration / 2, self.options)

        # repeat 8*M-1 times the pi pulse and free evolution of ps
        for itr_M in range(8 * self.M - 1):
            self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params[itr_M])
            self._free_evolution(ps, self.options)

        # perform the last pi pulse on X
        self._pulse(self.Ht, self.pi_pulse_duration, self.options, self.pulse_params[-1])

        if self.projection_pulse:
            self._free_evolution(ps / 2 - self.pi_pulse_duration / 2, self.options)
            self._pulse(self.Ht, self.pi_pulse_duration / 2, self.options, self.pulse_params[0])
        else:
            self._free_evolution(ps / 2, self.options)

        return self.rho

    def _get_pulse_profiles(self, tau=None):
        """
        Generates the pulse profiles for the XY8-M sequence for a given tau. The pulse profiles are stored in the pulse_profiles attribute of the object.

        Parameters
        ----------
        tau : float
            free evolution variable or pulse spacing for the  sequence
        """
        # check if tau is correctly defined
        if tau is None:
            tau = self.variable[-1]
        elif not isinstance(tau, (int, float)) or tau < self.pi_pulse_duration:
            raise ValueError("tau must be a positive real number larger than pi_pulse_duration")

        self.pulse_profiles = []

        # add the first pi/2 pulse on X axis
        if isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params[0],
                ]
            )
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(0, self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params[0],
                    ]
                    for i in range(len(self.H1))
                ]
            )

        t0 = self.pi_pulse_duration / 2
        ps = tau - self.pi_pulse_duration

        # add the first free evolution
        self.pulse_profiles.append([None, [t0, t0 + ps / 2 - self.pi_pulse_duration / 2], None, None])
        t0 += ps / 2 - self.pi_pulse_duration / 2

        # add pulses and free evolution M-1 times
        for itr_M in range(8 * self.M - 1):
            # add a pi pulse
            if isinstance(self.H1, Qobj):
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params[itr_M % 8],
                    ]
                )
            elif isinstance(self.H1, list):
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params[itr_M % 8],
                        ]
                        for i in range(len(self.H1))
                    ]
                )
            t0 += self.pi_pulse_duration

            # add a free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps], None, None])
            t0 += ps

        # add another pi pulse
        if isinstance(self.H1, Qobj):
            self.pulse_profiles.append(
                [
                    self.H1,
                    np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                    self.pulse_shape,
                    self.pulse_params[0],
                ]
            )
        elif isinstance(self.H1, list):
            self.pulse_profiles.append(
                [
                    [
                        self.H1[i],
                        np.linspace(t0, t0 + self.pi_pulse_duration, self.time_steps),
                        self.pulse_shape[i],
                        self.pulse_params[0],
                    ]
                    for i in range(len(self.H1))
                ]
            )
        t0 += self.pi_pulse_duration

        if self.projection_pulse:
            # add the last free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps / 2 - self.pi_pulse_duration / 2], None, None])
            t0 += ps / 2 - self.pi_pulse_duration / 2

            if isinstance(self.H1, Qobj):
                # add the last pi/2 pulse
                self.pulse_profiles.append(
                    [
                        self.H1,
                        np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                        self.pulse_shape,
                        self.pulse_params[0],
                    ]
                )
            elif isinstance(self.H1, list):
                # add the first pi/2 pulse
                self.pulse_profiles.append(
                    [
                        [
                            self.H1[i],
                            np.linspace(t0, t0 + self.pi_pulse_duration / 2, self.time_steps),
                            self.pulse_shape[i],
                            self.pulse_params[0],
                        ]
                        for i in range(len(self.H1))
                    ]
                )

            t0 += self.pi_pulse_duration / 2
        else:
            # add the last free evolution
            self.pulse_profiles.append([None, [t0, t0 + ps / 2], None, None])
            t0 += ps / 2

        # set the total_time attribute to the total time of the pulse sequence
        self.total_time = t0

    def plot_pulses(
        self, figsize=(6, 6), xlabel=None, ylabel="Expectation Value", title="Pulse Profiles", tau=None
    ):
        """
        Overwrites the plot_pulses method of the parent class in order to first generate the pulse profiles for the XY8-M sequence for a given tau and then plot them.

        Parameters
        ----------
        tau : float
            Free evolution time for the  sequence. Contrary to the run method, the free evolution must be a single number in order to plot the pulse profiles.
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        self._get_pulse_profiles(tau)
        # call the plot_pulses method of the parent class
        super().plot_pulses(figsize, xlabel, ylabel, title)



# TODO: add electric and crystal stress hamiltonians for NV

"""
This module contains NV class, which is a subclass of QSys.
"""

import warnings

import numpy as np
import scipy.constants as cte
from qutip import Qobj, basis, fock_dm, jmat, qeye, tensor

from .qsys import QSys

gamma_e = cte.value("electron gyromag. ratio in MHz/T") * 1e-3  # MHz/mT
gamma_N14 = 3.077e-3
gamma_N15 = -4.316e-3


class NV(QSys):
    """
    NN class contains attributes and methods to simulate the nitrogen vacancy center in diamond.

    Attributes
    ----------
    B0 : float
        Magnetic field
    N : 15, 14, 0 or None
        Nitrogen isotope, or 0 for no nuclear spin
    units_B0 : str
        Units of the magnetic field (T, mT or G)
    theta : float
        Angle of the magnetic field with respect to the NV axis
    phi_r : float
        Azimutal angle of the magnetic field with the NV axis
    units_angles : str
        Units of the angles (deg or rad)
    temp : float or None
        Temperature
    units_temp : str
        Temperature units 'C' or 'K'
    energy_levels : list
        List of energy levels of the Hamiltonian
    MW_freqs : numpy.ndarray
        Microwave frequencies
    RF_freqs : numpy.ndarray
        RF frequencies
    MW_H1 : Qobj
        Microwave Hamiltonian
    RF_H1 : Qobj
        RF Hamiltonian

    Methods
    -------
    rho0_lowT
        Calculates the initial state of the system at low temperatures using the Boltzmann distribution
    _set_MW_freqs
        Sets the standard resonant microwave frequencies for the NV center corresponding to the electronic spin transitions
    _set_RF_freqs
        Sets the standard resonant RF frequencies for the NV center corresponding to the nuclear spin transitions
    _set_MW_H1
        Sets the standard microwave Hamiltonian for the NV center corresponding to the electronic spin transitions
    _set_RF_H1
        Sets the standard RF Hamiltonian for the NV center corresponding to the nuclear spin transitions
    _ZeroField
        Get the NV Hamiltonian term accounting for zero field splitting
    _ElectronZeeman
        Get the NV hamiltonian term accounting for the electron Zeeman effect
    _NuclearZeeman
        Get the NV hamiltonian term accounting for the nuclear (Nitrogen) Zeeman effect
    _HyperfineN
        Get the NV hamiltonian term accounting for the hyperfine coupling with Nitrogen
    _Quadrupole
        Get the quadrupole term
    add_spin
        Adds an extra spin to the NV system
    truncate
        Truncates the system to the given indexes

    Notes
    -----
    The NV class inherits the methods and attributes from the QSys class.
    """

    def __init__(
        self,
        B0,
        N,
        c_ops=None,
        units_B0=None,
        theta=0.0,
        phi_r=0.0,
        units_angles="deg",
        temp=None,
        units_temp="K",
        E=0,
    ):
        """
        Constructor for the NV class.
        Takes the nitrogen isotope, the magnetic field intensity and angles with the quantization axis as inputs and calculates the energy levels of the Hamiltonian.

        Parameters
        ----------
        B0 : float
            Magnetic field
        N : 15/14/0/None
            Nitrogen isotope, or 0 for no nuclear spin
        c_ops : list(Qobj)
            List of collapse operators
        units_B0 : str
            Units of the magnetic field (T, mT or G)
        theta : float
            Angle of the magnetic field with respect to the NV axis
        phi_r : float
            Angle of the magnetic field in the xy plane
        units_angles : str
            Units of the angles (deg or rad)
        temp : float
            Temperature
        units_temp : str
            Temperature units ('C'/'K')
        E : float
            Perpedicular component of the zero field splitting
        """
        if not isinstance(B0, (int, float)):
            raise TypeError(f"B0 must be a real number, got {B0}: {type(B0)}.")
        else:
            self.B0 = B0

        if units_B0 is None:
            warnings.warn(
                "No units for the magnetic field were given. The magnetic field will be considered in mT."
            )
        elif units_B0 == "T":
            self.B0 = B0 * 1e3
        elif units_B0 == "mT":
            pass
        elif units_B0 == "G":
            self.B0 = B0 * 1e-1
        else:
            raise ValueError(
                f"Invalid value for units_B0. Expected either 'G', 'mT' or 'T', got {units_B0}."
            )

        if not isinstance(theta, (int, float)) or not isinstance(phi_r, (int, float)):
            raise TypeError(
                f"Invalid type for theta or phi_r. Expected a float or int, got theta: {type(theta)}, phi_r: {type(phi_r)}."
            )
        elif units_angles == "deg":
            theta = np.deg2rad(theta)
            phi_r = np.deg2rad(phi_r)
        elif units_angles == "rad":
            pass
        else:
            raise ValueError(
                f"Invalid value for units_angles. Expected either 'deg' or 'rad', got {units_angles}."
            )

        if not isinstance(E, (int, float)):
            raise TypeError(f"E must be a real number, got {E}: {type(E)}.")
        else:
            self.E = E

        self.theta = theta
        self.phi_r = phi_r
        self.N = N

        # calculates the Hamiltonian for the given field and nitrogen isotope
        if N == 15:
            H0 = self._ZeroField() + self._ElectronZeeman() + self._HyperfineN() + self._NuclearZeeman()
            rho0 = tensor(fock_dm(3, 1), qeye(2)).unit()
            observable = tensor(fock_dm(3, 1), qeye(2))

        elif N == 14:
            H0 = (
                self._ZeroField()
                + self._ElectronZeeman()
                + self._HyperfineN()
                + self._NuclearZeeman()
                + self._Quadrupole()
            )
            rho0 = tensor(fock_dm(3, 1), qeye(3)).unit()
            observable = tensor(fock_dm(3, 1), qeye(3))

        elif N == 0 or N is None:
            H0 = self._ZeroField() + self._ElectronZeeman()
            rho0 = basis(3, 1)
            observable = fock_dm(3, 1)

        else:
            raise ValueError(f"Invalid value for Nitrogen isotope. Expected either 14 or 15, got {N}.")

        super().__init__(H0, rho0, c_ops, observable, units_H0="MHz")

        if temp is not None:
            self.rho0_lowT(temp, units_temp)

        self._set_MW_freqs()
        self._set_RF_freqs()
        self._set_MW_H1()
        self._set_RF_H1()

    def rho0_lowT(self, temp, units_temp="K"):
        """
        Calculates the initial state of the system at low temperatures using the Boltzmann distribution.
        At room temperatures and moderate fields, the initial state of the nuclear spins is simply an identity matrix.

        Parameters
        ----------
        T : float
            Temperature
        units_temp : str
            Units of the temperature (K or C)

        Returns
        -------
        rho0 : Qobj
            Initial state of the system
        """
        if units_temp == "K":
            pass
        elif units_temp == "C":
            temp += 273.15
        elif units_temp == "F":
            raise ValueError("'F' is not a valid unit for temperature, learn the metric system.")
        else:
            raise ValueError(f"Invalid value for units_temp. Expected either 'K' or 'C', got {units_temp}.")

        if not isinstance(temp, (int, float)) and temp > 0:
            raise ValueError("T must be a positive real number.")

        # a loop to find the |0,1/2> and |0,-1/2> states
        max_1 = 0
        max_2 = 0
        max_3 = 0
        index_1 = None
        index_2 = None
        index_3 = None
        # iterates over all the eigenstates and find the one closest related to the |0,1/2> and |0,-1/2> states
        for itr in range(len(self.eigenstates)):
            if self.N == 15:
                proj_1 = np.abs(self.eigenstates[itr].overlap(basis(6, 2)))
                proj_2 = np.abs(self.eigenstates[itr].overlap(basis(6, 3)))

            elif self.N == 14:
                proj_1 = np.abs(self.eigenstates[itr].overlap(basis(9, 3)))
                proj_2 = np.abs(self.eigenstates[itr].overlap(basis(9, 4)))
                proj_3 = np.abs(self.eigenstates[itr].overlap(basis(9, 5)))
                if proj_3 > max_3:
                    # if the projection is higher than the previous maximum, update the maximum and the index
                    max_3 = proj_3
                    index_3 = itr

            if proj_1 > max_1:
                max_1 = proj_1
                index_1 = itr
            if proj_2 > max_2:
                max_2 = proj_2
                index_2 = itr

        beta = -cte.h * 1e6 / (cte.Boltzmann * temp)

        if self.N == 15:
            # calculate the partition function based on the Hamiltonian eigenvalues
            Z = np.exp(beta * self.energy_levels[index_1]) + np.exp(beta * self.energy_levels[index_2])

            self.rho0 = tensor(
                fock_dm(3, 1),
                Qobj(
                    [
                        [np.exp(beta * self.energy_levels[index_1]), 0],
                        [0, np.exp(beta * self.energy_levels[index_2])],
                    ]
                )
                / Z,
            )

        elif self.N == 14:
            Z = (
                np.exp(beta * self.energy_levels[index_1])
                + np.exp(beta * self.energy_levels[index_2])
                + np.exp(beta * self.energy_levels[index_3])
            )

            self.rho0 = tensor(
                fock_dm(3, 1),
                Qobj(
                    [
                        [np.exp(beta * self.energy_levels[index_1]), 0, 0],
                        [0, np.exp(beta * self.energy_levels[index_2]), 0],
                        [0, 0, np.exp(beta * self.energy_levels[index_3])],
                    ]
                )
                / Z,
            )
        elif self.N == 0 or self.N is None:
            self.rho0 = basis(3, 1)
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _set_MW_freqs(self):
        """
        Sets the standard resonant microwave frequencies for the NV center corresponding to the electronic spin transitions.
        """
        if self.N == 15:
            f1 = (np.sum(self.energy_levels[2:4]) - np.sum(self.energy_levels[1:2])) / 2
            f2 = (np.sum(self.energy_levels[4:6]) - np.sum(self.energy_levels[1:2])) / 2
            self.MW_freqs = np.array([f1, f2])
        elif self.N == 14:
            f1 = (np.sum(self.energy_levels[3:6]) - np.sum(self.energy_levels[1:3])) / 3
            f2 = (np.sum(self.energy_levels[6:9]) - np.sum(self.energy_levels[1:3])) / 3
            self.MW_freqs = np.array([f1, f2])
        elif self.N == 0 or self.N is None:
            f1 = self.energy_levels[1]
            f2 = self.energy_levels[2]
            self.MW_freqs = np.array([f1, f2])
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _set_RF_freqs(self):
        """
        Sets the standard resonant RF frequencies for the NV center corresponding to the nuclear spin transitions.
        """
        if self.N == 15:
            f1 = self.energy_levels[1]
            f2 = self.energy_levels[3] - self.energy_levels[2]
            f3 = self.energy_levels[5] - self.energy_levels[4]
            self.RF_freqs = np.array([f1, f2, f3])
        # for the 14N isotope, the RF frequencies are more complicated as they need to respect the selection rule of Delta mI = +-1
        elif self.N == 14:
            # the order of the ms states changes above the GSLAC
            if self.B0 <= 102.5:
                f1 = self.energy_levels[2] - self.energy_levels[1]  # 0 -> -1 at ms=0
                f2 = self.energy_levels[2]  # 0 -> +1 at ms=0
                f3 = self.energy_levels[5] - self.energy_levels[3]  # 0 -> -1 at ms=-1
                f4 = self.energy_levels[5] - self.energy_levels[4]  # 0 -> +1 at ms=-1
                f5 = self.energy_levels[8] - self.energy_levels[7]  # 0 -> -1 at ms=+1
                f6 = self.energy_levels[8] - self.energy_levels[6]  # 0 -> +1 at ms=-1
            else:
                f1 = self.energy_levels[2]  # 0 -> -1 at ms=-1
                f2 = self.energy_levels[2] - self.energy_levels[1]  # 0 -> +1 at ms=-1
                f3 = self.energy_levels[5] - self.energy_levels[4]  # 0 -> -1 at ms=0
                f4 = self.energy_levels[5] - self.energy_levels[3]  # 0 -> +1 at ms=0
                f5 = self.energy_levels[8] - self.energy_levels[7]  # 0 -> -1 at ms=+1
                f6 = self.energy_levels[8] - self.energy_levels[6]  # 0 -> +1 at ms=-1

            self.RF_freqs = np.array([f1, f2, f3, f4, f5, f6])
        elif self.N == 0 or self.N is None:
            self.RF_freqs = None
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _set_MW_H1(self):
        """
        Sets the standard microwave Hamiltonian for the NV center corresponding to the electronic spin transitions.
        """
        if self.N == 15:
            self.MW_H1 = tensor(jmat(1, "x"), qeye(2)) * 2**0.5
        elif self.N == 14:
            self.MW_H1 = tensor(jmat(1, "x"), qeye(3)) * 2**0.5
        elif self.N == 0 or self.N is None:
            self.MW_H1 = tensor(jmat(1, "x")) * 2**0.5
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _set_RF_H1(self):
        """
        Sets the standard RF Hamiltonian for the NV center corresponding to the nuclear spin transitions.
        """
        if self.N == 15:
            self.RF_H1 = tensor(qeye(3), jmat(1 / 2, "x")) * 2
        elif self.N == 14:
            self.RF_H1 = tensor(qeye(3), jmat(1, "x")) * 2**0.5
        elif self.N == 0 or self.N is None:
            self.RF_H1 = qeye(3)
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _ZeroField(self):
        """Get the NV Hamiltonian term accounting for zero field splitting.

        Parameters
        ----------
        D : float
            Axial component of magnetic dipole-dipole interaction, by default 2.87e3 MHz (NV)
        E : float
            Non axial compononet, by default 0. Usually much (1000x) smaller than `D`

        Returns
        -------
        Zero Field Hamiltonian : Qobj
        """
        if self.N == 14:
            return tensor(
                2.87e3 * jmat(1, "z") ** 2 + self.E * (jmat(1, "x") ** 2 - jmat(1, "y") ** 2), qeye(3)
            )
        elif self.N == 15:
            return tensor(
                2.87e3 * jmat(1, "z") ** 2 + self.E * (jmat(1, "x") ** 2 - jmat(1, "y") ** 2), qeye(2)
            )
        elif self.N == 0:
            return 2.87e3 * jmat(1, "z") ** 2 + self.E * (jmat(1, "x") ** 2 - jmat(1, "y") ** 2)
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _ElectronZeeman(self):
        """
        Get the NV hamiltonian term accounting for the electron Zeeman effect.

        Returns
        -------
        Electron Zeeman Hamiltonian : Qobj
        """

        if self.N == 14:
            return tensor(
                gamma_e
                * self.B0
                * (
                    np.cos(self.theta) * jmat(1, "z")
                    + np.sin(self.theta) * np.cos(self.phi_r) * jmat(1, "x")
                    + np.sin(self.theta) * np.sin(self.phi_r) * jmat(1, "y")
                ),
                qeye(3),
            )
        elif self.N == 15:
            return tensor(
                gamma_e
                * self.B0
                * (
                    np.cos(self.theta) * jmat(1, "z")
                    + np.sin(self.theta) * np.cos(self.phi_r) * jmat(1, "x")
                    + np.sin(self.theta) * np.sin(self.phi_r) * jmat(1, "y")
                ),
                qeye(2),
            )
        elif self.N == 0 or self.N is None:
            return (
                gamma_e
                * self.B0
                * (
                    np.cos(self.theta) * jmat(1, "z")
                    + np.sin(self.theta) * np.cos(self.theta) * jmat(1, "x")
                    + np.sin(self.theta) * np.sin(self.phi_r) * jmat(1, "y")
                )
            )
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _NuclearZeeman(self):
        """
        Get the NV hamiltonian term accounting for the nuclear (Nitrogen) Zeeman effect.

        Returns
        -------
        Nuclear Zeeman Hamiltonian : Qobj
        """

        if self.N == 14:
            return -tensor(
                qeye(3),
                gamma_N14
                * self.B0
                * (
                    np.cos(self.theta) * jmat(1, "z")
                    + np.sin(self.theta) * np.cos(self.phi_r) * jmat(1, "x")
                    + np.sin(self.theta) * np.sin(self.phi_r) * jmat(1, "y")
                ),
            )
        elif self.N == 15:
            return -tensor(
                qeye(3),
                gamma_N15
                * self.B0
                * (
                    np.cos(self.theta) * jmat(1 / 2, "z")
                    + np.sin(self.theta) * np.cos(self.phi_r) * jmat(1 / 2, "x")
                    + np.sin(self.theta) * np.sin(self.phi_r) * jmat(1 / 2, "y")
                ),
            )
        elif self.N == 0 or self.N is None:
            return 0
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _HyperfineN(self):
        """
        Get the NV hamiltonian term accounting for the hyperfine coupling with Nitrogen.

        Returns
        -------
        Hyperfine Hamiltonian : Qobj
        """
        if self.N == 14:
            return -2.14 * tensor(jmat(1, "z"), jmat(1, "z")) - 2.7 * (
                tensor(jmat(1, "x"), jmat(1, "x")) + tensor(jmat(1, "y"), jmat(1, "y"))
            )
        elif self.N == 15:
            return +3.03 * tensor(jmat(1, "z"), jmat(1 / 2, "z")) + 3.65 * (
                tensor(jmat(1, "x"), jmat(1 / 2, "x")) + tensor(jmat(1, "y"), jmat(1 / 2, "y"))
            )
        if self.N == 0 or self.N is None:
            return 0
        else:
            raise ValueError(f"Invalid value for Nitrogen. Expected either 14 or 15, got {self.N}.")

    def _Quadrupole(self):
        """
        Get the quadrupole term

        Returns
        -------
        Quadrupole Hamiltonian : Qobj
        """
        if self.N == 14:
            return -5.01 * tensor(qeye(3), jmat(1, "z") ** 2)
        elif self.N == 15 or self.N == 0:
            return None
        else:
            raise ValueError(
                f"Invalid value for nitrogen isotope N. Expected either 14 or 15, got {self.N}."
            )

    def add_spin(self, H_spin):
        """
        Overwrites the parent class method by calling it and updating MW_H1 and RF_H1 attributes

        Parameters
        ----------
        H_spin : Qobj
            Hamiltonian of the extra spin
        """
        super().add_spin(H_spin)

        self.MW_H1 = tensor(self.MW_H1, qeye(self.dim_add_spin))
        self.RF_H1 = tensor(self.RF_H1, qeye(self.dim_add_spin))

    def truncate(self, mS=None, mI=None):
        """
        Overwrites the parent class method by calling it and updating MW_H1 and RF_H1 attributes.
        The indexes to be removed are calculated according to the mS and mI parameters.

        Parameters
        ----------
        indexes : list(int)
            List of indexes to remove in the system
        """
        if mS is None and mI is None:
            warnings.warn("No mS or mI parameters were given. The system will not be truncated.")
            return
        if mS != 1 and mS != 0 and mS != -1 and mS is not None:
            raise ValueError(f"Invalid value for mS. Expected either 1, 0 or -1, got {mS}.")
        if mI == 1 / 2 or mI == -1 / 2:
            warnings.warn(
                "mI should be either 1, 0 or -1 for the NV system. The 15N isotope is already a two-level system and can't be truncated."
            )
        elif mI != 1 and mI != 0 and mI != -1 and mI is not None:
            raise ValueError(f"Invalid value for mI. Expected either 1, 0 or -1, got {mI}.")

        # set the indexes to be removed and the dimensions of the new objects according to the mS and mI parameters
        if self.N == 0 or self.N is None:
            if mS == 1:
                indexes, dims = [0], [2]
            elif mS == 0:
                indexes, dims = [1], [2]
            elif mS == -1:
                indexes, dims = [2], [2]

        elif self.N == 15:
            if mS == 1:
                indexes, dims = [0, 1], [2, 2]
            elif mS == 0:
                indexes, dims = [2, 3], [2, 2]
            elif mS == -1:
                indexes, dims = [4, 5], [2, 2]

        elif self.N == 14:
            if mS == 1:
                indexes, dims = [0, 1, 2], [2]
            elif mS == 0:
                indexes, dims = [3, 4, 5], [2]
            elif mS == -1:
                indexes, dims = [6, 7, 8], [2]
            else:
                dims = [3]

            if mI == 1:
                indexes.extend([0, 3, 6])
                dims.append(2)
            elif mI == 0:
                indexes.extend([1, 4, 7])
                dims.append(2)
            elif mI == -1:
                indexes.extend([2, 5, 8])
                dims.append(2)
            else:
                dims.append(3)

        indexes = sorted(set(indexes))
        super().truncate(indexes)

        self.MW_H1 = Qobj(np.delete(np.delete(self.MW_H1.full(), indexes, axis=0), indexes, axis=1))
        self.RF_H1 = Qobj(np.delete(np.delete(self.RF_H1.full(), indexes, axis=0), indexes, axis=1))

        # corrrect the dimensions of the objects
        self.H0.dims = [dims, dims]
        self.MW_H1.dims = [dims, dims]
        self.RF_H1.dims = [dims, dims]

        if self.observable is not None:
            if isinstance(self.observable, Qobj):
                self.observable.dims = [dims, dims]
            elif isinstance(self.observable, list):
                for obs in self.observable:
                    obs.dims = [dims, dims]

        if self.rho0 is not None:
            if self.rho0.isket:
                if len(dims) == 1:
                    self.rho0.dims = [dims, [1]]
                elif len(dims) == 2:
                    self.rho0.dims = [dims, [1, 1]]
            else:
                self.rho0.dims = [dims, dims]

        if self.c_ops is not None:
            if isinstance(self.c_ops, Qobj):
                self.c_ops.dims = [dims, dims]
            elif isinstance(self.c_ops, list):
                for c_op in self.c_ops:
                    c_op.dims = [dims, dims]



# TODO: expand and test

"""
This module contains pulse shape functions to be used in the PulsedSim classes as part of the QuaCAAToo package.
"""

import numpy as np


def square_pulse(t, **pulse_params):
    """
    Square pulse envelope modulation

    Parameters
    ----------
    t : float
        Time parameter
    f_pulse : float
        Frequency of the pulse
    phi_t : float
        Phase of the pulse

    Returns
    -------
    float
        Value of the square pulse at time t
    """
    return np.cos(pulse_params["f_pulse"] * t + pulse_params["phi_t"])


def gaussian_pulse(t, **pulse_params):
    """
    Gaussian pulse envelope modulation

    Parameters
    ----------
    t : float
        Time parameter
    t_mid : float
        Middle point of the pulse
    sigma : float
        Width of the pulse
    f_pulse : float
        Frequency of the pulse
    phi_t : float
        Phase of the pulse

    Returns
    -------
    float
        Value of the Gaussian pulse at time t
    """
    return np.exp(-((t - pulse_params["t_mid"]) ** 2) / (2 * pulse_params["sigma"] ** 2)) * np.cos(
        pulse_params["f_pulse"] * t + pulse_params["phi_t"]
    )


def lorentzian_pulse(t, **pulse_params):
    """
    Lorentzian pulse envelope modulation

    Parameters
    ----------
    t : float
        Time parameter
    t_mid : float
        Middle point of the pulse
    gamma : float
        Width of the pulse
    f_pulse : float
        Frequency of the pulse
    phi_t : float
        Phase of the pulse

    Returns
    -------
    float
        Value of the Lorentzian pulse at time t
    """
    return (
        1
        / (1 + ((t - pulse_params["t_mid"]) / ["gamma"]) ** 2)
        * np.cos(pulse_params["f_pulse"] * t + pulse_params["phi_t"])
    )



"""
Collection of in built fit models and functions.

The predefined model classes take a (fit) function and implement a guess subroutine for intelligent initial values.

The general mechanism is the same as that of lmfit. So a lot of functions have been taken from there.
To create additional fit models, simply instantiate an lmfit model with the target fit function.

Refer to the documentation of analysis.run_fit for usage details.
"""

# ruff: noqa: F401
import numpy as np
from lmfit import Model
from lmfit.models import (
    ConstantModel,
    ExponentialModel,
    GaussianModel,
    LinearModel,
    LorentzianModel,
    SineModel,
    update_param_vals,
)

############################################## Rabi - Periodic Oscillation ##############################################


def _guess_sin(data, x):
    """
    Internal helper function to provide decent first guesses for sinusoidal functions.

    Parameters
    ----------
    data: array_like
        The dependent variable over which fit is to be performed
    x: array_like
        The independent variable in the fitting

    Returns
    -------
    amp: float
        Guessed amplitude
    frequency: float
        Guessed frequency
    """
    y = data - data.mean()
    frequencies = np.fft.fftfreq(len(x), abs(x[-1] - x[0]) / (len(x) - 1))
    fft = abs(np.fft.fft(y))
    argmax = abs(fft).argmax()
    amp = 2.0 * fft[argmax] / len(fft)
    frequency = 2 * np.pi * abs(frequencies[argmax])

    return amp, frequency


def fit_rabi(x, amp=1, Tpi=10, phi=0, offset=0):
    """
    Fit a cosine function to Rabi oscillations.

    Parameters
    ----------
    x : array_like
        Time values.
    amp : float
        Amplitude of the cosine function.
    Tpi : float
        Pi-pulse duration (half the period of the cosine function).
    phi : float
        Phase of the cosine function.
    offset : float
        Offset of the cosine function.
    """
    return amp * np.cos(np.pi * x / Tpi + phi) + offset


class RabiModel(Model):
    """
    Modified from lmfit's SineModel to include an offset as well

    Takes the same parameters as the fit_rabi function.
    """

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update({"prefix": prefix, "nan_policy": nan_policy, "independent_vars": independent_vars})
        super().__init__(fit_rabi, **kwargs)
        self._set_paramhints_prefix()

    def guess(self, data, x, **kwargs):
        offset = data.mean()
        amp, frequency = _guess_sin(data, x)
        data = data - data.mean()

        # try shifts in the range [0, 2*pi) and take the one with best residual
        shift_guesses = np.linspace(0, 2 * np.pi, 11, endpoint=False)
        errors = [
            np.linalg.norm(
                self.eval(x=x, amp=amp, Tpi=np.pi / frequency, phi=shift_guess, offset=offset) - data
            )
            for shift_guess in shift_guesses
        ]
        phi = shift_guesses[np.argmin(errors)]
        pars = self.make_params(amp=amp, Tpi=np.pi / frequency, phi=phi, offset=offset)
        return update_param_vals(pars, self.prefix, **kwargs)


################################################ Exponential Decay ####################################################


def _guess_exp(data, x):
    """
    Internal helper function to provide decent first guesses for exponential functions.

    Parameters
    ----------
    data: array_like
        The dependent variable over which fit is to be performed
    x: array_like
        The independent variable in the fitting

    Returns
    -------
    coeff: array
        An array with amplitude and decay time as its elements.
    """
    y = np.log(np.abs(data))
    result = np.polynomial.Polynomial.fit(x, y, 1)
    coeff = result.convert().coef
    coeff[0] = np.exp(coeff[0])
    coeff[1] = -1 / coeff[1]
    return coeff


def fit_exp_decay(x, amp=1, Tc=1, offset=0):
    """
    Fit a simple exponential decay function.

    Parameters
    ----------
    x : array_like
        Time values.
    amp : float
        Amplitude of the exponential decay.
    Tc : float
        Decay time constant.
    offset : float
        Offset of the exponential decay.
    """
    return amp * np.exp(-x / Tc) + offset


class ExpDecayModel(Model):
    """
    Modified from lmfit's ExponentialModel to include an offset as well

    Takes the same parameters as the fit_exp_decay function.
    """

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update({"prefix": prefix, "nan_policy": nan_policy, "independent_vars": independent_vars})
        super().__init__(fit_exp_decay, **kwargs)

    def guess(self, data, x, **kwargs):
        coeff = _guess_exp(data, x)
        pars = self.make_params(amp=coeff[0], Tc=coeff[1], offset=data[-1])
        return update_param_vals(pars, self.prefix, **kwargs)


######################################################### Rabi with Exp Decay ####################################################


def fit_rabi_decay(x, amp=1, Tpi=10, phi=0, offset=0, Tc=1):
    """
    Fit a cosine function with exponential decay to Rabi oscillations.

    Parameters
    ----------
    x : array_like
        Time values.
    amp : float
        Amplitude of the cosine function.
    Tpi : float
        Pi-pulse duration (half the period of the cosine function).
    phi : float
        Phase of the cosine function.
    offset : float
        Offset of the cosine function.
    Tc : float
        Decay time constant.
    """
    return amp * np.cos(np.pi * x / Tpi + phi) * np.exp(-x / Tc) + offset


class RabiDecayModel(Model):
    """
    Analogous to RabiModel, modulated with an exponential decay.

    Takes the same parameters as the fit_rabi_decay function.
    """

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update({"prefix": prefix, "nan_policy": nan_policy, "independent_vars": independent_vars})
        super().__init__(fit_rabi_decay, **kwargs)

    def guess(self, data, x, **kwargs):
        offset = data.mean()
        amp, frequency = _guess_sin(data, x)
        data = data - data.mean()

        coeff = _guess_exp(data, x)

        shift_guesses = np.linspace(0, 2 * np.pi, 11, endpoint=False)
        errors = [
            np.linalg.norm(
                self.eval(x=x, amp=amp, Tpi=np.pi / frequency, phi=shift_guess, offset=offset, Tc=coeff[1])
                - data
            )
            for shift_guess in shift_guesses
        ]
        phi = shift_guesses[np.argmin(errors)]

        pars = self.make_params(amp=amp, Tpi=np.pi / frequency, phi=phi, offset=offset, Tc=coeff[1])

        return update_param_vals(pars, self.prefix, **kwargs)


def fit_exp_decay_n(x, A=1, C=0, Tc=1, n=1):
    """
    Fit an exponential decay function with power n.

    Parameters
    ----------
    x : array_like
        Time values.
    A : float
        Amplitude of the exponential decay.
    C : float
        Offset of the exponential decay.
    Tc : float
        Decay time constant.
    n : float
        Power of the exponential decay.
    """
    return A * np.exp(-((x / Tc) ** n)) + C


####################################################### Hahn Modulation ##############################################################


def fit_hahn_mod(x, A, B, C, f1, f2):
    """
    Fit a Hahn echo with modulation function with 2 frequencies.

    Parameters
    ----------
    x : array_like
        Time values.
    A : float
        Amplitude of the echo.
    B : float
        Amplitude of the modulation.
    C : float
        Offset of the echo.
    f1 : float
        First modulation frequency.
    f2 : float
        Second modulation frequency.
    """
    return (A - B * np.sin(2 * np.pi * f1 * x / 2) ** 2 * np.sin(2 * np.pi * f2 * x / 2) ** 2) + C


def fit_hahn_mod_decay(x, A, B, C, f1, f2, Tc, n):
    """
    Fit a Hahn echo with modulation function with 2 frequencies and exponential decay.

    Parameters
    ----------
    x : array_like
        Time values.
    A : float
        Amplitude of the echo.
    B : float
        Amplitude of the modulation.
    C : float
        Offset of the echo.
    f1 : float
        First modulation frequency.
    f2 : float
        Second modulation frequency.
    Tc : float
        Decay time constant.
    n : float
        Power of the exponential decay.
    """
    return (
        np.exp(-((x / Tc) ** n))
        * (A - B * np.sin(2 * np.pi * f1 * x / 2) ** 2 * np.sin(2 * np.pi * f2 * x / 2) ** 2)
        + C
    )


####################################################### Lorentzian and sinc ##############################################################


def fit_lorentz(x, A, gamma, f0, C):
    """
    Fit a Lorentzian peak.

    Parameters
    ----------
    x : array_like
        Frequency values.
    A : float
        Amplitude of the peak.
    gamma : float
        Width of the peak.
    f0 : float
        Central requency of the peak.
    C : float
        Offset of the peak.
    """
    return C - A * (gamma**2) / ((x - f0) ** 2 + gamma**2)


def fit_two_lorentz(x, A1=1, A2=1, gamma1=0.1, gamma2=0.1, f01=2.87e3, f02=2.87e3, C=0):
    """
    Fit two symmetric Lorentzian peaks.

    Parameters
    ----------
    x : array_like
        Frequency values.
    A1 : float
        Amplitude of the first peak.
    A2 : float
        Amplitude of the second peak.
    gamma1 : float
        Width of the first peak.
    gamma2 : float
        Width of the second peak.
    f01 : float
        Central frequency of the first peak.
    f02 : float
        Central frequency of the second peak.
    C : float
        Offset of the peaks.
    """
    return C + fit_lorentz(x, A1, gamma1, f01, 0) + fit_lorentz(x, A2, gamma2, f02, 0)


def fit_two_lorentz_sym(x, A=1, gamma=1, f_mean=2.87e3, f_delta=1, C=0):
    """
    Fit two symmetric Lorentzian peaks.

    Parameters
    ----------
    x : array_like
        Frquency values.
    A : float
        Amplitude of the peaks.
    gamma : float
        Width of the peaks.
    f_mean : float
        Mean frequency of the peaks.
    f_delta : float
        Frequency difference between the peaks.
    C : float
        Offset of the peaks.
    """
    return (
        C
        + fit_lorentz(x, A, gamma, f_mean - f_delta / 2, 0)
        + fit_lorentz(x, A, gamma, f_mean + f_delta / 2, 0)
    )


def fit_sinc2(x, A, gamma, f0, C):
    """
    Fit a sinc function.

    Parameters
    ----------
    x : array_like
        Frequency values.
    A : float
        Amplitude of the sinc function.
    gamma : float
        Width of the sinc function.
    f0 : float
        Central frequency of the sinc function.
    C : float
        Offset of the sinc function.
    """
    return (
        C
        - A
        * gamma**2
        / (gamma**2 + (x - f0) ** 2)
        * np.sin((gamma**2 + (x - f0) ** 2) ** 0.5 / gamma / 2 * np.pi) ** 2
    )


def fit_two_sinc2_sym(x, A, gamma, f_mean, f_delta, C):
    """
    Fit two symmetric sinc functions.

    Parameters
    ----------
    x : array_like
        Frequency values.
    A : float
        Amplitude of the sinc functions.
    gamma : float
        Width of the sinc functions.
    f_mean : float
        Mean frequency of the sinc functions.
    f_delta : float
        Frequency difference between the sinc functions.
    C : float
        Offset of the sinc functions.
    """
    return (
        C + fit_sinc2(x, A, gamma, f_mean - f_delta / 2, 0) + fit_sinc2(x, A, gamma, f_mean + f_delta / 2, 0)
    )


def fit_five_sinc2(
    x, A1, A2, A3, A4, A5, gamma1, gamma2, gamma3, gamma4, gamma5, f01, f02, f03, f04, f05, C
):
    """
    Fit two symmetric sinc functions.

    Parameters
    ----------
    x : array_like
        Frequency values.
    A : float
        Amplitude of the sinc functions.
    gamma : float
        Width of the sinc functions.
    f_mean : float
        Mean frequency of the sinc functions.
    f_delta : float
        Frequency difference between the sinc functions.
    C : float
        Offset of the sinc functions.
    """
    return (
        C
        + fit_sinc2(x, A1, gamma1, f01, 0)
        + fit_sinc2(x, A2, gamma2, f02, 0)
        + fit_sinc2(x, A3, gamma3, f03, 0)
        + fit_sinc2(x, A4, gamma4, f04, 0)
        + fit_sinc2(x, A5, gamma5, f05, 0)
    )



"""
This module contains the ExpData class as part of the QuaCCAToo package.
"""

import matplotlib.pyplot as plt
import numpy as np


class ExpData:
    """
    Class to load experimental data and perform basic data processing.

    Attributes
    ----------
    variable : np.ndarray
        Array containing the variable data
    results : np.ndarray or list of np.ndarray
        Array or list of arrays containing the results data
    variable_name : str
        Name of the variable
    result_name : str
        Name of the results

    Methods
    -------
    subtract_results_columns:
        Subtracts the results of the negative column from the positive column
    offset_correction:
        Substracts a background value from the results
    rescale_correction:
        Multiplies the results by a rescale value
    poly_baseline_correction:
        Performs a polynomial baseline correction to the results
    plot_exp_data:
        Plots the experimental data
    """

    def __init__(
        self,
        file_path,
        variable_column=0,
        results_columns=1,
        variable_name="Time",
        result_name="Expectation Value",
        plot=False,
        figsize=(6, 4),
        figtitle="Experimental Data",
        **loadtxt_kwargs,
    ):
        """
        Constructor of the ExpData class.
        It loads experimental data from a file and sets the variable and results attributes according with the specified column arguments.

        Parameters
        ----------
        file_path  : str
            Path to the file containing the experimental data
        variable_column : int
            Column index of the variable
        results_columns : int or list of int
            Column index of the results
        variable_name : str
            Name of the variable
        result_name : str
            Name of the results
        plot : bool
            Boolean indicating whether to plot the experimental data or not
        figsize : tuple
            Size of the figure for the plot
        figtitle : str
            Title of the figure for the plot
        **loadtxt_kwargs : dict
            Additional arguments for the np.loadtxt function
        """
        if not isinstance(file_path, str):
            raise ValueError("file_path must be a string")

        if not isinstance(variable_column, int):
            raise ValueError("variable_column must be an integer")

        # the results columns needs to be an integer or a list of integers
        if not isinstance(results_columns, int) and not (
            isinstance(results_columns, list) and all(isinstance(col, int) for col in results_columns)
        ):
            raise ValueError("results_columns must be an integer or a list of two integers")

        if not isinstance(variable_name, str) or not isinstance(result_name, str):
            raise ValueError("variable_name and result_name must be strings")

        if not isinstance(loadtxt_kwargs, dict):
            raise ValueError("loadtxt_kwargs must be a dictionary for the np.loadtxt function")

        # loads experimental data from a file with the specified arguments
        exp_data = np.loadtxt(file_path, **loadtxt_kwargs)

        # sets the results and variable attributes of the ExpData object
        self.variable = exp_data[:, variable_column]

        if isinstance(results_columns, int):
            self.results = exp_data[:, results_columns]
        else:
            self.results = [exp_data[:, column] for column in results_columns]

        self.variable_name = variable_name
        self.result_name = result_name

        if not isinstance(plot, bool):
            raise ValueError("plot must be a boolean")
        # plots the experimental data
        elif plot:
            self.plot_exp_data(figsize=figsize, figtitle=figtitle)

    def subtract_results_columns(
        self, pos_col=0, neg_col=1, plot=False, figsize=(6, 4), figtitle="Subtracted Expt. Data"
    ):
        """
        Overwrites the results attribute substracting the results of the negative column from the positive column.

        Parameters
        ----------
        pos_col: int
            Index of the positive column
        neg_col: int
            Index of the negative column
        plot: bool
            Boolean indicating whether to plot the experimental data or not
        figsize: tuple
            Size of the figure for the plot
        figtitle: str
            Title of the figure for the plot
        """
        if not isinstance(self.results[pos_col], np.ndarray) or not isinstance(
            self.results[neg_col], np.ndarray
        ):
            raise ValueError(f"pos_col={pos_col} and neg_col={neg_col} where not found in the results.")

        self.results = self.results[pos_col] - self.results[neg_col]

        if not isinstance(plot, bool):
            raise ValueError("plot must be a boolean")
        elif plot:
            self.plot_exp_data(figsize=figsize, figtitle=figtitle)

    def offset_correction(
        self, background_value, plot=False, figsize=(6, 4), figtitle="Expt. Data with Offset Correction"
    ):
        """
        Overwrites the results attribute substracting the background value from the results.

        Parameters
        ----------
        background_value : int or float
            Value to be substracted from the results
        plot : bool
            Boolean indicating whether to plot the experimental data or not
        figsize : tuple
            Size of the figure for the plot
        figtitle : str
            Title of the figure for the plot
        """
        if not isinstance(background_value, (int, float)):
            raise ValueError("background_value must be a number.")

        if isinstance(self.results, np.ndarray):
            self.results = self.results - background_value
        elif isinstance(self.results, list) and all(
            isinstance(result, np.ndarray) for result in self.results
        ):
            self.results = [result - background_value for result in self.results]
        else:
            raise ValueError("Results must be a numpy array or a list of numpy arrays")

        if not isinstance(plot, bool):
            raise ValueError("plot must be a boolean")
        elif plot:
            self.plot_exp_data(figsize=figsize, figtitle=figtitle)

    def rescale_correction(
        self, rescale_value, plot=False, figsize=(6, 4), figtitle="Expt. Data with Rescale Correction"
    ):
        """
        Overwrites the results attribute multiplying the results by the rescale value.

        Parameters
        ----------
        rescale_value : int or float
            Value to be multiplied by the results
        plot : bool
            Boolean indicating whether to plot the experimental data or not
        figsize : tuple
            Size of the figure for the plot
        figtitle : str
            Title of the figure for the plot
        """
        if not isinstance(rescale_value, (int, float)):
            raise ValueError("rescale_value must be a number.")

        if isinstance(self.results, np.ndarray):
            self.results = self.results * rescale_value
        elif isinstance(self.results, list) and all(
            isinstance(result, np.ndarray) for result in self.results
        ):
            self.results = [result * rescale_value for result in self.results]
        else:
            raise ValueError("Results must be a numpy array or a list of numpy arrays")

        if not isinstance(plot, bool):
            raise ValueError("plot must be a boolean")
        elif plot:
            self.plot_exp_data(figsize=figsize, figtitle=figtitle)

    def poly_base_correction(
        self,
        x_start=None,
        x_end=None,
        poly_order=2,
        plot=False,
        figsize=(6, 4),
        figtitle="Expt. Data with Polynomial Baseline Correction",
    ):
        """
        Overwrites the results attribute performing a polynomial baseline correction.
        The baseline is fitted to the data between x_start and x_end, representing the start and end of the xaxis index.

        Parameters
        ----------
        x_start : int or list of int
            Start index of the x axis for the baseline fit
        x_end : int or list of int
            End index of the x axis for the baseline fit
        poly_order : int
            Order of the polynomial to fit the baseline
        plot : bool
            Boolean indicating whether to plot the experimental data or not
        figsize : tuple
            Size of the figure for the plot
        figtitle : str
            Title of the figure for the plot

        """
        # check all variables
        if x_start is None:
            x_start = 0
        elif not isinstance(x_start, int) and not (
            isinstance(x_start, list) and all(isinstance(x, int) for x in x_start)
        ):
            raise ValueError("x_start must be a integer index or a list of integer indexes.")

        if x_end is None:
            x_end = -1
        elif not isinstance(x_end, int) and not (
            isinstance(x_end, list) and all(isinstance(x, int) for x in x_end)
        ):
            raise ValueError("x_end must be a integer index or a list of integer indexes.")

        if not isinstance(poly_order, int):
            raise ValueError("poly_order must be an integer.")

        # crops the x and y axis for performing the baseline fit
        if isinstance(x_start, int) and isinstance(x_end, int):
            baseline_xaxis = self.variable[x_start:x_end]
            baseline_yaxis = self.results[x_start:x_end]
        elif isinstance(x_start, list) and isinstance(x_end, list) and len(x_start) == len(x_end):
            baseline_xaxis = np.concatenate(
                [self.variable[x_start[i] : x_end[i]] for i in range(len(x_start))]
            )
            baseline_yaxis = np.concatenate(
                [self.results[x_start[i] : x_end[i]] for i in range(len(x_start))]
            )
        else:
            raise ValueError("x_start and x_end must int or a list of the same length.")

        if isinstance(self.results, np.ndarray):
            poly_fit = np.polyfit(baseline_xaxis, baseline_yaxis, poly_order)
            self.results -= np.polyval(poly_fit, self.variable)

        elif isinstance(self.results, list) and all(
            isinstance(result, np.ndarray) for result in self.results
        ):
            poly_fit = [
                np.polyfit(baseline_xaxis[i], baseline_yaxis[i], poly_order)
                for i in range(len(baseline_xaxis))
            ]
            self.results = [
                self.results[i] - np.polyval(poly_fit[i], self.variable) for i in range(len(self.results))
            ]

        if not isinstance(plot, bool):
            raise ValueError("plot must be a boolean")
        elif plot:
            self.plot_exp_data(figsize=figsize, figtitle=figtitle)

    def plot_exp_data(self, figsize=(6, 4), figtitle="Experimental Data"):
        """
        Plots the experimental data.

        Parameters
        ----------
        figsize : tuple
            Size of the figure for the plot
        figtitle : str
            Title of the figure for the plot
        """
        if not (isinstance(figsize, tuple) or len(figsize) == 2):
            raise ValueError("figsize must be a tuple of two positive floats")

        if not isinstance(figtitle, str):
            raise ValueError("figtitle must be a string")

        fig, ax = plt.subplots(1, 1, figsize=figsize)

        # check if the results is a list of results or a single result
        if isinstance(self.results, np.ndarray):
            ax.scatter(self.variable, self.results, lw=2, alpha=0.7, label="Observable", s=15)

        elif isinstance(self.results, list) and all(
            isinstance(result, np.ndarray) for result in self.results
        ):
            for itr in range(len(self.results)):
                ax.scatter(self.variable, self.results[itr], label=f"Observable {itr}", alpha=0.7, s=15)

        else:
            raise ValueError("Results must be a numpy array or a list of numpy arrays")

        # set the x-axis limits to the variable of the experiment
        ax.set_xlim(self.variable[0], self.variable[-1])

        ax.set_xlabel(self.variable_name)
        ax.set_ylabel(self.result_name)
        ax.legend(loc="upper right", bbox_to_anchor=(1.2, 1))
        ax.set_title(figtitle)
